"""This module contains the GridObject class.
"""
import copy as cp
from typing import Tuple, List

import numpy as np
import numpy.typing as npt

import matplotlib.pyplot as plt
from matplotlib import colors
from mpl_toolkits.axes_grid1 import make_axes_locatable

from scipy.ndimage import (
    convolve,
    median_filter,
    generic_filter,
    grey_erosion,
    grey_dilation,
    distance_transform_edt
)
from scipy.signal import wiener

from rasterio import CRS, Affine
from rasterio.coords import BoundingBox
from rasterio.warp import reproject, transform_bounds
from rasterio.enums import Resampling

# pylint: disable=no-name-in-module
from . import _grid, _morphology  # type: ignore
from .interface import validate_alignment

__all__ = ['GridObject']


class GridObject():
    """A class containing all information of a Digital Elevation Model (DEM).
    """

    def __init__(self) -> None:
        """Initialize a GridObject instance.
        """
        # path to file
        self.path = ''
        # name of DEM
        self.name = ''

        # raster metadata
        self.z: npt.NDArray = np.empty((), dtype=np.float32)

        self.cellsize = 0.0  # in meters if crs.is_projected == True

        # georeference
        self.bounds = None
        self.transform = Affine.identity()
        self.georef = None

    @property
    def shape(self):
        """Tuple of grid dimensions
        """
        return self.z.shape

    @property
    def rows(self):
        """The size of the first dimension of the grid
        """
        return self.z.shape[0]

    @property
    def columns(self):
        """The size of the second dimension of the grid
        """
        return self.z.shape[1]

    @property
    def dims(self):
        """The dimensions of the grid in the correct order for libtopotoolbox
        """
        if self.z.flags.c_contiguous:
            return (self.columns, self.rows)

        if self.z.flags.f_contiguous:
            return (self.rows, self.columns)

        raise TypeError(
            "Grid is not stored as a contiguous row- or column-major array")

    @property
    def extent(self):
        """The bounding box of the grid in the order needed for plotting

        Returns
        -------
        tuple
            The bounding box in the order (left, right, bottom, top)
        """
        if self.bounds:
            return (self.bounds.left, self.bounds.right, self.bounds.bottom, self.bounds.top)

        return (-0.5, self.columns-0.5, self.rows-0.5, -0.5)

    @property
    def coordinates(self):
        """Coordinate arrays for the DEM

        Returns
        -------
        X,Y : tuple of ndarrays
            The two returned arrays are of the same shape as the
        DEM. The first contains the coordinates of each pixel in the
        horizontal dimension, and the second contains the coordinates
        in the vertical dimension.
        """
        x, y = np.meshgrid(np.arange(self.columns), np.arange(self.rows))
        return self.transform * (x, y)

    def astype(self, dtype):
        """Copy of the GridObject, cast to specified type

        Parameters
        ----------
        dtype: str or np.dtype
            The numpy data type to which the GridObject is cast

        Returns
        -------
        GridObject
            A copy of the original GridObject with the given data type

        Example
        -------
        >>> import topotoolbox
        >>> import numpy as np
        >>> dem = topotoolbox.load_dem('perfectworld')
        >>> dem = dem.astype(np.float32)
        """
        result = GridObject()
        result.path = self.path
        result.name = self.name
        result.z = self.z.astype(dtype, copy=True)

        result.cellsize = self.cellsize

        result.bounds = self.bounds
        result.transform = self.transform
        result.georef = self.georef

        return result

    def reproject(self,
                  georef: 'CRS',
                  resolution: 'float | None' = None,
                  resampling: 'Resampling' = Resampling.bilinear):
        """Reproject GridObject to a new coordinate system.

        Parameters
        ----------
        georef : rasterio.CRS
            Target coordinate system
        resolution : float, optional
            Target resolution.
            If one is not provided, a resolution that approximately
            matches that of the source coordinate system will be used.
        resampling : rasterio.enums.Resampling, optional
            Resampling method.
            The default is bilinear resampling.

        Returns
        -------
        GridObject
            The reprojected data.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> from rasterio import CRS
           >>> dem = topotoolbox.load_dem('kunashiri')
           >>> dem = dem.reproject(CRS.from_epsg(32655), resolution = 30)
           >>> _= dem.plot(cmap="terrain")
           >>> plt.show()
        """
        dst = GridObject()

        z, dst.transform = reproject(
            self.z,
            src_transform=self.transform,
            src_crs=self.georef,
            dst_transform=None,  # Let rasterio derive the transform for us
            dst_crs=georef,
            dst_nodata=np.nan,
            dst_resolution=resolution,
            resampling=resampling,
        )
        # reproject gives us a 3D array, we want the first band.
        dst.z = np.zeros_like(self.z, shape=z.shape[1:3])
        dst.z[:, :] = z[0, :, :]

        dst.georef = georef

        # Get cellsize from transform in case we did not specify one
        if dst.transform is not None:
            dst.cellsize = abs(dst.transform[0])

        if self.bounds:
            dst.bounds = BoundingBox(
                *transform_bounds(self.georef, dst.georef, *self.bounds))

        return dst

    def resample(self,
                 resolution: float,
                 resampling: 'Resampling' = Resampling.bilinear):
        """Resample GridObject to a new resolution.

        Parameters
        ----------
        resolution : float
            Target resolution in the same units as the current cellsize.
        resampling : rasterio.enums.Resampling, optional
            Resampling method.
            The default is bilinear resampling.

        Returns
        -------
        GridObject
            The resampled data.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> dem_90m = dem.resample(resolution=90)
           >>> im = dem_90m.plot(cmap="terrain")
           >>> plt.show()
        """
        return self.reproject(self.georef, resolution, resampling)

    def fillsinks(self,
                  bc: 'np.ndarray | GridObject | None' = None,
                  hybrid: bool = True) -> 'GridObject':
        """Fill sinks in the digital elevation model (DEM).

        Parameters
        ----------
        bc : ndarray or GridObject, optional
            Boundary conditions for sink filling. `bc` should be an array
            of np.uint8 that matches the shape of the DEM. Values of 1
            indicate pixels that should be fixed to their values in the
            original DEM and values of 0 indicate pixels that should be
            filled.
        hybrid: bool, optional
            Should hybrid reconstruction algorithm be used? Defaults to True. Hybrid
            reconstruction is faster but requires additional memory be allocated
            for a queue.

        Returns
        -------
        GridObject
            The filled DEM.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> filled_dem = dem.fillsinks()
           >>> _= filled_dem.plot(cmap='terrain')
           >>> plt.show()
        """
        dem = self.z.astype(np.float32)
        output = np.zeros_like(dem)

        restore_nans = False

        if bc is None:
            bc = np.ones_like(dem, dtype=np.uint8)
            bc[1:-1, 1:-1] = 0  # Set interior pixels to 0

            nans = np.isnan(dem)
            dem[nans] = -np.inf
            bc[nans] = 1  # Set NaNs to 1
            restore_nans = True

        if not validate_alignment(self, bc):
            err = ("The shape of the provided boundary conditions does not "
                   f"match the shape of the DEM. {self.shape}")
            raise ValueError(err)from None

        bc = np.asarray(bc, dtype=np.uint8,
                        order = 'F' if dem.flags.f_contiguous else 'C')

        if hybrid:
            queue = np.zeros_like(dem, dtype=np.int64)
            _grid.fillsinks_hybrid(output, queue, dem, bc, self.dims)
        else:
            _grid.fillsinks(output, dem, bc, self.dims)

        if restore_nans:
            dem[nans] = np.nan
            output[nans] = np.nan

        result = cp.copy(self)
        result.z = output

        return result

    def identifyflats(
            self, raw: bool = False, output: list[str] | None = None) -> tuple:
        """Identifies flats and sills in a digital elevation model (DEM).

        Parameters
        ----------
        raw : bool, optional
            If True, returns the raw output grid as np.ndarray.
            Defaults to False.
        output : list of str, optional
            List of strings indicating desired output types. Possible values
            are 'sills', 'flats'. Order of inputs in list are irrelevant,
            first entry in output will always be sills.
            Defaults to ['sills', 'flats'].

        Returns
        -------
        tuple
            A tuple containing copies of the DEM with identified
            flats and/or sills.

        Notes
        -----
        Flats are identified as 1s, sills as 2s, and presills as 5s
        (since they are also flats) in the output grid.
        Only relevant when using raw=True.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> flats, sills = dem.identifyflats()
           >>> _= flats.plot(cmap='terrain')
           >>> plt.show()
        """

        # Since having lists as default arguments can lead to problems, output
        # is initialized with None by default and only converted to a list
        # containing default output here:
        if output is None:
            output = ['sills', 'flats']

        dem = np.asarray(self, dtype=np.float32)
        output_grid = np.zeros_like(dem, dtype=np.int32)

        _grid.identifyflats(output_grid, dem, self.dims)

        if raw:
            return (output_grid,)

        result = []
        if 'flats' in output:
            flats = cp.copy(self)
            flats.z = np.zeros_like(flats.z)
            flats.z = np.where((output_grid & 1) == 1, 1, flats.z)
            result.append(flats)

        if 'sills' in output:
            sills = cp.copy(self)
            sills.z = np.zeros_like(sills.z)
            sills.z = np.where((output_grid & 2) == 2, 1, sills.z)
            result.append(sills)

        return tuple(result)

    def excesstopography(
            self, threshold: "float | int | np.ndarray | GridObject" = 0.2,
            method: str = 'fsm2d',) -> 'GridObject':
        """
        Compute the two-dimensional excess topography using the specified method.

        Parameters
        ----------
        threshold : float, int, GridObject, or np.ndarray, optional
            Threshold value or array to determine slope limits, by default 0.2.
            If a float or int, the same threshold is applied to the entire DEM.
            If a GridObject or np.ndarray, it must match the shape of the DEM.
        method : str, optional
            Method to compute the excess topography, by default 'fsm2d'.
            Options are:

            - 'fsm2d': Uses the fast sweeping method.
            - 'fmm2d': Uses the fast marching method.

        Returns
        -------
        GridObject
            A new GridObject with the computed excess topography.

        Raises
        ------
        ValueError
            If `method` is not one of ['fsm2d', 'fmm2d'].
            If `threshold` is an np.ndarray and doesn't match the shape of the DEM.
        TypeError
            If `threshold` is not a float, int, GridObject, or np.ndarray.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> excess = dem.excesstopography(threshold=0.3, method='fsm2d')
           >>> _= excess.plot(cmap='terrain')
           >>> plt.show()
        """

        if method not in ['fsm2d', 'fmm2d']:
            err = (f"Invalid method '{method}'. Supported methods are" +
                   " 'fsm2d' and 'fmm2d'.")
            raise ValueError(err) from None

        dem = np.asarray(self, dtype=np.float32)

        if isinstance(threshold, (float, int)):
            threshold_slopes = np.full_like(dem, threshold)
        elif isinstance(threshold, GridObject):
            threshold_slopes = np.asarray(threshold,
                                          order='F' if self.z.flags.f_contiguous else 'C',
                                          dtype=np.float32)
        elif isinstance(threshold, np.ndarray):
            threshold_slopes = np.asarray(threshold,
                                          order='F' if self.z.flags.f_contiguous else 'C',
                                          dtype=np.float32)
        else:
            err = "Threshold must be a float, int, GridObject, or np.ndarray."
            raise TypeError(err) from None

        if not validate_alignment(dem, threshold_slopes):
            err = "Threshold array must have the same shape as the DEM."
            raise ValueError(err) from None

        excess = np.zeros_like(dem)
        cellsize = self.cellsize

        if method == 'fsm2d':
            _grid.excesstopography_fsm2d(
                excess, dem, threshold_slopes, cellsize, self.dims)

        elif method == 'fmm2d':
            heap = np.zeros_like(dem, dtype=np.int64)
            back = np.zeros_like(dem, dtype=np.int64)

            _grid.excesstopography_fmm2d(excess, heap, back, dem,
                                         threshold_slopes, cellsize,
                                         self.dims)

        result = cp.copy(self)
        result.z = excess

        return result

    def filter(self, method: str = 'mean', kernelsize: int = 3) -> 'GridObject':
        """The function filter is a wrapper around various image filtering
        algorithms. Only filters with rectangular kernels of uneven
        dimensions are supported.

        Parameters
        ----------
        method : str, optional
            Which method will be used to filter the DEM: ['mean', 'average',
            'median', 'sobel', 'scharr', 'wiener', 'std'], by default 'mean'
        kernelsize : int, optional
            Size of the kernel that will be applied. Note that ['sobel',
            'scharr'] require that the kernelsize is 3, by default 3

        Returns
        -------
        GridObject
            The filtered DEM as a GridObject.

        Raises
        ------
        ValueError
            If the kernelsize does not match the requirements of this function
            or the selected method is not implemented in the function.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> scharr = dem.filter(method='scharr', kernelsize=3)
           >>> _= scharr.plot(cmap='terrain')
           >>> plt.show()
        """

        valid_methods = ['mean', 'average', 'median',
                         'sobel', 'scharr', 'wiener', 'std']

        if method in ['mean', 'average']:
            kernel = np.ones((kernelsize, kernelsize)) / kernelsize**2
            filtered = convolve(self.z, kernel, mode='nearest')

        elif method in ['median']:
            filtered = median_filter(self.z, size=kernelsize, mode='reflect')

        elif method in ['sobel', 'scharr']:
            if kernelsize != 3:
                arr = f"The method '{method}' only works with a 3x3 kernel'."
                raise ValueError(arr) from None

            if method == 'sobel':
                ky = np.array([[1, 2, 1], [0, 0, 0], [-1, -2, -1]])
            else:  # 'scharr'
                ky = np.array([[3, 10, 3], [0, 0, 0], [-3, - 10, - 3]])

            kx = ky.T
            filtered = np.hypot(
                convolve(self.z, ky, mode='nearest'),
                convolve(self.z, kx, mode='nearest'))

        elif method in ['wiener']:
            filtered = wiener(self.z, mysize=kernelsize)

        elif method in ['std']:
            # This solution is based on this thread:
            # https://stackoverflow.com/questions/19518827/what-is-the-python
            # -equivalent-for-matlabs-stdfilt-function
            filtered = generic_filter(self.z, np.std, size=kernelsize)
            factor = np.sqrt(kernelsize**2 / (kernelsize**2 - 1))
            np.multiply(filtered, factor, out=filtered)

        else:
            err = (f"Argument 'method={method}' has to be"
                   f"one of {valid_methods}.")
            raise ValueError(err) from None

        # Keep NaNs like they are in self.z
        filtered[np.isnan(self.z)] = np.nan

        result = cp.copy(self)
        result.z = filtered
        return result

    def gradient8(self, unit: str = 'tangent', multiprocessing: bool = True):
        """Compute the gradient of a digital elevation model (DEM) using an
        8-direction algorithm.

        Parameters
        ----------
        unit : str, optional
          The unit of the gradient to be calculated. Options are:
          - 'tangent' : Calculate the gradient as a tangent (default).
          - 'radian'  : Calculate the gradient in radians.
          - 'degree'  : Calculate the gradient in degrees.
          - 'sine'    : Calculate the gradient as the sine of the angle.
          - 'percent' : Calculate the gradient as a percentage.
        multiprocessing : bool, optional
          If True, use multiprocessing for computation. Default is True.

        Returns
        -------
        GridObject
          A new GridObject with the calculated gradient.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> grad = dem.gradient8()
           >>> _= grad.plot(cmap='terrain')
           >>> plt.show()
        """

        if multiprocessing:
            use_mp = 1
        else:
            use_mp = 0

        dem = np.asarray(self.z, dtype=np.float32)
        output = np.zeros_like(dem)

        _grid.gradient8(output, dem, self.cellsize, use_mp, self.dims)
        result = cp.copy(self)

        if unit == 'radian':
            output = np.arctan(output)
        elif unit == 'degree':
            output = np.rad2deg(np.arctan(output))
        elif unit == 'sine':
            output = np.sin(np.arctan(output))
        elif unit == 'percent':
            output = output * 100.0

        result.z = output

        return result

    def curvature(self, ctype='profc', meanfilt=False) -> 'GridObject':
        """curvature returns the second numerical derivative (curvature) of a
        digital elevation model. By default, curvature returns the profile
        curvature (profc).

        Parameters
        ----------
        ctype : str, optional
            What type of curvature will be computed, by default 'profc'
            - 'profc' : profile curvature [m^(-1)],
            - 'planc' : planform curvature [m^(-1))],
            - 'tangc' : tangential curvature [m^(-1)],
            - 'meanc' : mean curvature [m^(-1)],
            - 'total' : total curvature [m^(-2)]
        meanfilt : bool, optional
            True if a mean filter will be applied before comuting the
            curvature, by default False

        Returns
        -------
        GridObject
            A GridObject storing the computed values.

        Raises
        ------
        ValueError
            If wrong ctype has been used.

        Examples
        --------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('tibet')
           >>> curv = dem.curvature()
           >>> _= curv.plot(cmap='terrain')
           >>> plt.show()
        """

        if meanfilt:
            kernel = np.ones((3, 3)) / 9
            dem = convolve(self.z, kernel, mode='nearest')
        else:
            dem = self.z

        kernel_fx = np.array(
            [[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]]) / (6 * self.cellsize)
        kernel_fy = np.array(
            [[1, 1, 1], [0, 0, 0], [-1, -1, -1]]) / (6 * self.cellsize)

        fx = convolve(dem, kernel_fx, mode='nearest')
        fy = convolve(dem, kernel_fy, mode='nearest')

        kernel_fxx = np.array(
            [[1, -2, 1], [1, -2, 1], [1, -2, 1]]) / (3 * self.cellsize**2)
        kernel_fyy = kernel_fxx.T
        kernel_fxy = np.array(
            [[-1, 0, 1], [0, 0, 0], [1, 0, -1]]) / (4 * self.cellsize**2)

        fxx = convolve(dem, kernel_fxx, mode='nearest')
        fyy = convolve(dem, kernel_fyy, mode='nearest')
        fxy = convolve(dem, kernel_fxy, mode='nearest')

        epsilon = 1e-10
        if ctype == 'profc':
            denominator = (fx**2 + fy**2) * (1 + fx **
                                             2 + fy**2)**(3/2) + epsilon
            curvature = - (fx**2 * fxx + 2 * fx * fy * fxy + fy **
                           2 * fyy) / denominator
        elif ctype == 'tangc':
            denominator = (fx**2 + fy**2) * (1 + fx **
                                             2 + fy**2)**(1/2) + epsilon
            curvature = - (fy**2 * fxx - 2 * fx * fy * fxy + fx **
                           2 * fyy) / denominator
        elif ctype == 'planc':
            denominator = (fx**2 + fy**2)**(3/2) + epsilon
            curvature = - (fy**2 * fxx - 2 * fx * fy * fxy + fx **
                           2 * fyy) / denominator
        elif ctype == 'meanc':
            denominator = 2 * (fx**2 + fy**2 + 1)**(3/2) + epsilon
            curvature = -((1 + fy ** 2) * fxx - 2 * fxy * fx * fy +
                          (1 + fx ** 2) * fyy) / denominator
        elif ctype == 'total':
            curvature = fxx**2 + 2 * fxy**2 + fyy**2
        else:
            raise ValueError(
                "Invalid curvature type. Must be one of: 'profc', 'planc',"
                "'tangc', 'meanc', 'total'.")

        # Keep NaNs like they are in dem
        curvature[np.isnan(dem)] = np.nan

        result = cp.copy(self)
        result.z = curvature
        return result

    def dilate(
            self, size: tuple | None = None, footprint: np.ndarray | None = None,
            structure: np.ndarray | None = None) -> 'GridObject':
        """A simple wrapper around th scipy.ndimage.grey_dilation function,
        that also handles NaNs in the input GridObject. Either size, footprint
        or structure has to be passed to this function. If nothing is provided,
        the function will raise an error.

        size : tuple of ints, optional
            A tuple of ints containing the shape of the structuring element.
            Only needed if neither footprint nor structure is provided. Will
            result in a full and flat structuring element.
            Defaults to None
        footprint : np.ndarray of ints, optional
            A array defining the footprint of the erosion operation.
            Non-zero elements define the neighborhood over which the erosion
            is applied. Defaults to None
        structure : np.ndarray of ints, optional
            A array defining the structuring element used for the erosion.
            This defines the connectivity of the elements. Defaults to None

        Returns
        -------
        GridObject
            A GridObject storing the computed values.
        Raises
        ------
        ValueError
            If size, structure and footprint are all None.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> dilate = dem.dilate(size=10)
           >>> _= dilate.plot(cmap='terrain')
           >>> plt.show()
        """

        if size is None and structure is None and footprint is None:
            err = ("Dilate requires a structuring element to be specified."
                   " Use the size argument for a full and flat structuring"
                   " element (equivalent to a a minimum filter) or the"
                   " structure and footprint arguments to specify"
                   " a non-flat structuring element.")
            raise ValueError(err) from None

        # Replace NaN values with inf
        dem = self.z.copy()
        dem[np.isnan(dem)] = -np.inf

        dilated = grey_dilation(
            input=dem, size=size, structure=structure, footprint=footprint)

        # Keep NaNs like they are in dem
        dilated[np.isnan(self.z)] = np.nan

        result = cp.copy(self)
        result.z = dilated
        return result

    def erode(
            self, size: tuple | None = None, footprint: np.ndarray | None = None,
            structure: np.ndarray | None = None) -> 'GridObject':
        """Apply a morphological erosion operation to the GridObject. Either
        size, footprint or structure has to be passed to this function. If
        nothing is provided, the function will raise an error.

        Parameters
        ----------
        size : tuple of ints
            A tuple of ints containing the shape of the structuring element.
            Only needed if neither footprint nor structure is provided. Will
            result in a full and flat structuring element.
            Defaults to None
        footprint : np.ndarray of ints, optional
            A array defining the footprint of the erosion operation.
            Non-zero elements define the neighborhood over which the erosion
            is applied. Defaults to None
        structure : np.ndarray of ints, optional
            A array defining the structuring element used for the erosion.
            This defines the connectivity of the elements. Defaults to None

        Returns
        -------
        GridObject
            A GridObject storing the computed values.

        Raises
        ------
        ValueError
            If size, structure and footprint are all None.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> eroded = dem.erode(size = (3,3))
           >>> _= eroded.plot(cmap='terrain')
           >>> plt.show()
        """

        if size is None and structure is None and footprint is None:
            err = ("Erode requires a structuring element to be specified."
                   " Use the size argument for a full and flat structuring"
                   " element (equivalent to a a minimum filter) or the"
                   " structure and footprint arguments to specify"
                   " a non-flat structuring element.")
            raise ValueError(err) from None

        # Replace NaN values with inf
        dem = self.z.copy()
        dem[np.isnan(dem)] = np.inf

        eroded = grey_erosion(
            dem, size=size, structure=structure, footprint=footprint)

        # Keep NaNs like they are in dem
        eroded[np.isnan(self.z)] = np.nan

        result = cp.copy(self)
        result.z = eroded
        return result

    def evansslope(
            self, partial_derivatives: bool = False, mode: str = 'nearest',
            modified: bool = False) -> 'GridObject' | Tuple['GridObject', 'GridObject']:
        """Evans method fits a second-order polynomial to 3x3 subgrids. The
        parameters of the polynomial are the partial derivatives which are
        used to calculate the surface slope = sqrt(Gx**2 + Gy**2).

        Evans method approximates the surface by regression surfaces.
        Gradients are thus less susceptible to noise in the DEM.

        Parameters
        ----------
        mode : str, optional
            The mode parameter determines how the input DEM is extended
            beyond its boundaries: ['reflect', 'constant', 'nearest', 'mirror',
            'wrap', 'grid-mirror', 'grid-constant', 'grid-wrap']. See
            scipy.ndimage.convolve for more information, by default 'nearest'
        modified : bool, optional
            If True, the surface is weakly smoothed before gradients are
            calculated (see Shary et al., 2002), by default False
        partial_derivatives : bool, optional
            If True, both partial derivatives [fx, fy] will be returned as
            GridObjects instead of just the evansslope, by default False

        Returns
        -------
        GridObject
            A GridObject containing the computed evansslope data.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> slope = dem.evansslope()
           >>> _= slope.plot(cmap='terrain')
           >>> plt.show()
        """
        dem = self.z.copy()
        # NaN replacement not optional since convolve can't handle NaNs
        indices = distance_transform_edt(
            np.isnan(dem), return_distances=False, return_indices=True)
        dem = dem[tuple(indices)]

        if modified:
            kernel = np.array([[0, 1, 0], [1, 41, 1], [0, 1, 0]])/45
            dem = convolve(dem, kernel, mode=mode)

        kx = np.array(
            [[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]])/(6*self.cellsize)
        fx = convolve(dem, kx, mode=mode)
        # kernel for dz/dy
        ky = np.array(
            [[1, 1, 1], [0, 0, 0], [-1, -1, -1]])/(6*self.cellsize)
        fy = convolve(dem, ky, mode=mode)

        if partial_derivatives:
            result_kx = cp.copy(self)
            result_ky = cp.copy(self)
            result_kx.z = kx
            result_ky.z = ky
            return result_kx, result_ky

        slope = np.sqrt(fx**2 + fy**2)
        slope[np.isnan(self.z)] = np.nan

        result = cp.copy(self)
        result.z = slope
        return result

    def aspect(self, classify: bool = False) -> 'GridObject':
        """Aspect returns the slope exposition of each cell in a digital
        elevation model in degrees. In contrast to the second output of
        gradient8 which returns the steepest slope direction, aspect
        returns the angle of the slope.

        Parameters
        ----------
        classify : bool, optional
            directions are classified according to the scheme proposed by
            Gomez-Plaza et al. (2001), by default False

        Returns
        -------
        GridObject
            A GridObject containing the computed aspect data.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> aspect = dem.aspect()
           >>> _= aspect.plot(cmap='terrain')
           >>> plt.show()
        """

        grad_y, grad_x = np.gradient(self.z, edge_order=2)
        aspect: np.ndarray = np.arctan2(-grad_x, grad_y)
        aspect = np.degrees(aspect)
        aspect = np.mod(aspect, 360)

        if classify:
            aspclass = np.array([1, 3, 5, 7, 8, 6, 4, 2])
            aspedges = aspect // 45
            aspedges = aspedges.astype(np.int8)

            aspect = aspclass[aspedges]
            aspect = aspect.astype(np.int8)

        result = cp.copy(self)
        result.z = aspect
        return result

    def prominence(self, tolerance: float, use_hybrid=True) -> Tuple:
        """This function calculates the prominence of peaks in a DEM. The
        prominence is the minimal amount one would need to descend from a peak
        before being able to ascend to a higher peak. The function uses image
        reconstruct (see function imreconstruct) to calculate the prominence.
        It may take a while to run for large DEMs. The algorithm iteratively
        finds the next higher prominence and stops if the prominence is less
        than the tolerance, the second input parameter to the function.

        Parameters
        ----------
        tolerance : float
            The minimum tolerance for the second to last found peak. (meters)
            Will always find one peak.
        use_hybrid : bool, optional
            If True, use the hybrid reconstruction algorithm. Defaults to True.

        Returns
        -------
        Tuple[np.ndarray, Tuple]
            A Tuple containing a ndarray storing the computed
            prominence and the geographic coordinates of the peaks,
            stored as a tuple of ndarray.

        Examples
        --------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> from matplotlib.colors import ListedColormap
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> prom, (x, y) = dem.prominence(tolerance=90)
           >>> _ = plt.subplot()
           >>> _ = dem.plot_hs(cmap=ListedColormap([0.9, 0.9, 0.9]), exaggerate=3)
           >>> _ = plt.scatter(x, y, prom, alpha=0.6, edgecolor='black')
           >>> plt.show()
        """
        dem = np.nan_to_num(self.z)
        p = np.full_like(dem, np.min(dem))

        prominence: List[float] = []
        indices = []

        queue = np.zeros_like(dem, dtype=np.int64)

        while not prominence or prominence[-1] > tolerance:
            diff = dem - p
            prominence.append(np.max(diff))

            # By default argmax returns indices into the row-major
            # flattened array even if the array is not
            # row-major. However, unravel_index unravels by default in
            # row-major order, so the resulting pair of indices are
            # valid regardless of the memory order.
            indices.append(np.unravel_index(np.argmax(diff), self.shape))

            p[indices[-1]] = dem[indices[-1]]
            if use_hybrid:
                _morphology.reconstruct_hybrid(p, queue, dem, self.dims)
            else:
                _morphology.reconstruct(p, dem, self.dims)

        prominence_array = np.array(prominence)
        indices_array = np.array(indices)
        indices_array = indices_array[:, [1, 0]]  # swap columns 0 and 1
        # transpose to get (x, y) instead of (y, x)
        indices_array = self.transform * indices_array.T
        return prominence_array, indices_array

    def hillshade(self,
                  azimuth: float = 315.0,
                  altitude: float = 60.0,
                  exaggerate: float = 1.0,
                  fused=True):
        """Compute a hillshade of a digital elevation model

        Parameters
        ----------
        azimuth : float
            The azimuth angle of the light source measured in degrees
            clockwise from north. Defaults to 315 degrees.
        altitude : float
            The altitude angle of the light source measured in degrees
            above the horizon. Defaults to 60 degrees.
        exaggerate : float
            The amount of vertical exaggeration. Increase to emphasize
            elevation differences in flat terrain. Defaults to 1.0
        fused : bool, optional
            If true, use the fused hillshade computation in
            libtopotoolbox, which requires less memory but can be
            slightly slower. If you have a small DEM, and are
            repeatedly creating hillshades consider
            setting to False for increased performance. Defaults to True.

        Returns
        -------
        GridObject
            A GridObject containing the resulting hillshade data

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> hillshade = dem.hillshade()
           >>> _= hillshade.plot(cmap='gray')
           >>> _= dem.plot(cmap='terrain', alpha=0.2)
           >>> plt.show()
        """

        z = np.asarray(self, dtype=np.float32)
        h = np.zeros_like(z)

        # Computing the azimuth angle is a bit tricky
        gt = self.transform

        # Remove the translation from the geotransform. It is not
        # needed to rotate the coordinate system, but it makes things
        # harder to work with.
        gt = gt.translation(-gt.xoff, -gt.yoff)*gt

        if self.z.flags.f_contiguous:
            # If the array is column-major, we need to swap the x and
            # y coordinates, which is achieved with a matrix like
            #
            # [[0  -1],
            #  [-1  0]]
            #
            # which we can construct as a permutation followed by a
            # 180 degree rotation.
            gt = gt.rotation(180) * gt.permutation() * gt

        # This is the /east/ component of the azimuth vector
        sx = np.sin(np.deg2rad(azimuth))
        # This is the /north/ component of the azimuth vector
        sy = np.cos(np.deg2rad(azimuth))

        # Apply the inverse of the geotransform to convert from
        # geographic coordinates to image coordinates.
        dx, dy = ~gt * (sx, sy)

        # And retrieve the azimuth angle.
        azimuth_radians = np.arctan2(dy, dx)

        # NOTE(wkearn): This angle is then immediately used within
        # hillshade to compute vector components again. It would be
        # somewhat more efficient and numerically stable to work
        # directly with the vectors. libtopotoolbox's hillshade should
        # probably take a vector rather than an angle.
        #
        # See Inigo Quilez (2013). Avoiding trigonometry
        # (https://iquilezles.org/articles/noacos/) for an argument
        # against using angles in graphics APIs.

        altitude_radians = np.deg2rad(altitude)

        if fused:
            _grid.hillshade_fused(h, exaggerate * z,
                                  azimuth_radians, altitude_radians,
                                  self.cellsize, self.dims)
        else:
            nx = np.zeros_like(z)
            ny = np.zeros_like(z)
            _grid.hillshade(h, nx, ny, exaggerate * z,
                            azimuth_radians, altitude_radians,
                            self.cellsize, self.dims)

        result = cp.copy(self)
        result.z = h
        return result

    def info(self) -> None:
        """Prints all variables of a GridObject.
        """
        print(f"name: {self.name}")
        print(f"path: {self.path}")
        print(f"rows: {self.rows}")
        print(f"cols: {self.columns}")
        print(f"cellsize: {self.cellsize}")
        print(f"bounds: {self.bounds}")
        print(f"transform: {self.transform}")
        if self.georef is not None and self.georef.is_projected:
            print(f"coordinate system (Projected): {self.georef}")
        elif self.georef is not None and self.georef.is_geographic:
            print(f"coordinate system (Geographic): {self.georef}")
        else:
            print(f"coordinate system: {self.georef}")
        print(f"maximum z-value: {np.nanmax(self.z)}")
        print(f"minimum z-value: {np.nanmin(self.z)}")

    def plot(self, ax=None, extent=None, **kwargs):
        """Plot the GridObject

        Parameters
        ----------
        ax: matplotlib.axes.Axes, optional
            The axes in which to plot the GridObject. If no axes
            are given, the current axes are used.

        extent: floats (left, right, bottom, top), optional
            The bounding box used to set the axis limits. If no extent
            is supplied, defaults to self.extent, which plots the
            GridObject in geographic coordinates.

        **kwargs
            Additional keyword arguments are forwarded to
            matplotlib.axes.Axes.imshow

        Returns
        -------
        matplotlib.image.AxesImage
            The image constructed by imshow

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> _= dem.plot(cmap='terrain')
           >>> plt.show()
        """
        if ax is None:
            ax = plt.gca()

        if extent is None:
            extent = self.extent

        return ax.imshow(self.z, extent=extent, **kwargs)

    def plot_hs(self, ax=None,
                elev=None,
                azimuth=315, altitude=60, exaggerate=1,
                filter_method=None, filter_size=3,
                cmap='terrain', norm=None,
                blend_mode='soft',
                extent=None,
                colorbar = False,
                **kwargs):
        """Plot a shaded relief map of the GridObject

        Parameters
        ----------
        ax: matplotlib.axes.Axes, optional
            The axes in which to plot the GridObject. If no axes
            are given, the current axes are used.
        elev: GridObject, optional
            The digital elevation model used for shading. If no DEM is
            provided, the data GridObject is also used for shading.
        azimuth: float, optional
            The azimuth angle of the light source in degrees from
            North. Defaults to 315 degrees.
        altitude: float, optional
            The altitude angle of the light source in degrees above
            the horizon. Defaults to 60 degrees.
        exaggerate: float, optional
            The amount of vertical exaggeration to apply to the
            elevation. Defaults to 1.
        filter_method: 'str', optional
            The method used to filter the DEM before computing the
            hillshade. The data GridObject is not filtered. This
            should be one of the methods provided by
            `GridObject.filter`. Defaults to None, which does not
            apply a filter.
        filter_size: int, optional
            The size of the filter kernel in pixels. Defaults to 3.
        cmap: colors.Colormap or str or None
            The colormap to use in coloring the data. Defaults to
            'terrain'.
        norm: colors.Normalize, optional
            The normalization method that scales the color data to the
            [0,1] interval. Defaults to a linear scaling from the
            minimum of the data to the maximum.
        blend_mode: {'multiply', 'overlay', 'soft'}, optional
            The algorithm used to combine the shaded elevation with
            the data. Defaults to 'soft'.
        extent: floats (left, right, bottom, top), optional
            The bounding box used to set the axis limits. If no extent
            is supplied, defaults to self.extent
        **kwargs
            Additional keyword arguments are forwarded to
            matplotlib.axes.Axes.imshow

        Returns
        -------
        matplotlib.image.AxesImage
            The image constructed by imshow

        Raises
        ------
        TypeError
            The `elev` argument is not a GridObject

        ValueError
            The `elev` argument is not the same shape as the data

        ValueError
            A `blend_mode` other than 'multiply', 'overlay' or 'soft' is provided.

        ValueError
            The `filter_method` or `filter_size` arguments are not
            accepted by `GridObject.filter`.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('perfectworld')
           >>> _= dem.plot_hs(exaggerate=dem.cellsize)
           >>> plt.show()
        """
        if ax is None:
            ax = plt.gca()

        if elev is None:
            shade = self
        elif isinstance(elev, GridObject):
            if not validate_alignment(self, elev):
                err = "elev GridObject must have the same shape as the GridObject."
                raise ValueError(err) from None
            shade = elev
        else:
            err = "elev must be a GridObject"
            raise TypeError(err) from None

        if filter_method is not None:
            shade = shade.filter(method=filter_method, kernelsize=filter_size)

        h = shade.hillshade(azimuth, altitude, exaggerate)
        cmap = plt.get_cmap(cmap)

        if norm is None:
            norm = colors.Normalize(vmin=np.nanmin(
                self.z), vmax=np.nanmax(self.z))

        base = cmap(norm(self.z))
        top = np.expand_dims(np.clip(h, 0, 1), 2)
        if blend_mode == "multiply":
            rgb = base * top
        elif blend_mode == "overlay":
            rgb = np.where(base < 0.5, 2*base*top, 1 - 2*(1-base)*(1-top))
        elif blend_mode == "soft":
            rgb = (1 - 2*top)*base**2 + 2 * top * base
        else:
            raise ValueError("blend_mode not supported") from None

        if extent is None:
            extent = self.extent

        if colorbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), cax=cax)

        return ax.imshow(np.clip(rgb, 0, 1), extent=extent, **kwargs)

    def plot_surface(self, ax=None, **kwargs):
        """Plot DEM as a 3D surface

        Parameters
        ----------
        ax: matplotlib.axes.Axes, optional
            The axes in which to plot the GridObject. If no axes
            are given, the current axes are used.

        **kwargs
            Additional keyword arguments are forwarded to
            matplotlib.axes.Axes3D.plot_surface.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fig = plt.figure()
           >>> ax = fig.add_subplot(111, projection='3d')
           >>> _= dem.plot_surface(ax=ax)
           >>> ax.set_aspect('equal')
           >>> _= ax.set_zticks([0,np.nanmax(dem)])
           >>> plt.show()
        """
        if ax is None:
            ax = plt.gca()

        x, y = self.coordinates

        return ax.plot_surface(x, y, self.z, **kwargs)

    def shufflelabel(self, seed=None):
        """Randomize the labels of a GridObject

        This function is helpful when plotting drainage basins. It will work with
        any kind of data, but is most useful when given ordinal data such as an
        integer-valued GridObject.

        Parameters
        ----------
        seed: optional

          The seed used to generate the random permutation of labels.

          The seed is passed directly to `numpy.random.default_rng`__.

          .. __:
             https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng

        Returns
        -------
        GridObject
          A grid identical to the input, but with randomly reassigned labels.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('bigtujunga')
           >>> fd = topotoolbox.FlowObject(dem)
           >>> D = fd.drainagebasins()
           >>> _= D.shufflelabel().plot(cmap="Pastel1",interpolation="nearest")
           >>> plt.show()
        """
        result = cp.copy(self)

        u, indices = np.unique(self, return_inverse=True)
        rng = np.random.default_rng(seed)
        result.z = np.reshape(rng.permutation(u)[indices], self.shape)

        return result

    def duplicate_with_new_data(self, data: np.ndarray) -> 'GridObject':
        """Duplicate a GridObject with different data

        This function is helpful when one wants to create a GridObject from
        a numpy array with the exact same properties (e.g. georef, ...) but
        different data

        Parameters
        ----------
        data: np.ndarray

          The new data (needs to be in the same shape than the current GridObject)

        Returns
        -------
        GridObject
          A grid identical to the input, but with new data.

        Example
        -------
        >>> import topotoolbox
        >>> dem = topotoolbox.load_dem('perfectworld')
        >>> new_dem = dem.duplicate_with_new_data(np.zeros(dem.shape))
        """
        rows, columns = data.shape

        if self.columns != columns or self.rows != rows:
            raise ValueError("Both GridObjects have to be the same size.")

        result = cp.deepcopy(self)

        result.z = np.array(data, copy=True)

        return result

    def zscore(self, dtype=None):
        """Returns the z-score for each element of GridObject such that
        all values are centered to have mean 0 and scaled to have
        standard deviation 1.

        Parameters
        ----------
        dtype
            The data type of the resulting array. zscore uses
            np.float64 internally for precision and will convert back
            to the input GridObject's data type by default. If the
            additional precision is necessary, pass np.float64.

        Returns
        -------
        GridObject
            A GridObject containing the z-scores of the input GridObject.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('tibet')
           >>> dem_zscore = dem.zscore()
           >>> _= dem_zscore.plot()
           >>> plt.show()
        """
        result = cp.copy(self)
        z = (self.z - np.nanmean(self.z, dtype=np.float64)) / np.nanstd(self.z, dtype=np.float64)
        result.z = np.asarray(z, dtype=dtype if dtype else self.z.dtype)
        return result

    def crop(self, left: float | int, right: float | int,
               top: float | int, bottom: float | int,
               mode: str) -> 'GridObject':
        """Crop the Gridobject by specifying new boundaries.

        Supports three input modes (percent, coordinate, pixel) to define
        the crop region. Method of crop has to be chosen by using the mode
        argument. In case of reversed boundaries, automatically swaps them to
        ensure the crop region is valid. The resulting grid will have a new
        transform and bounds based on the specified boundaries.

        Parameters
        ----------
        left : float or int
            Left boundary in one of three modes:
            - Percent: 0.0 to 1.0 (relative to grid width)
            - Coordinate: Within grid's horizontal bounds
            - Pixel: Column index (0 to grid width-1)
        right : float or int
            Right boundary (same modes as `left`).
        top : float or int
            Top boundary in one of three modes:
            - Percent: 0.0-1.0 (relative to grid height)
            - Coordinate: Within grid's vertical bounds
            - Pixel: Row index (0 to grid height-1)
        bottom : float or int
            Bottom boundary (same modes as `top`).
        mode : str
            The mode of the crop operation. Can be 'percent', 'coordinate',
            or 'pixel'.

        Returns
        -------
        GridObject
            Cropped grid with updated transform, bounds, and data.

        Raises
        ------
        ValueError
            If boundaries are not in a consistent valid mode.

        Example
        -------
        .. plot::

           >>> import topotoolbox
           >>> import matplotlib.pyplot as plt
           >>> dem = topotoolbox.load_dem('tibet')
           >>> new_dem = dem.crop(0.6, 0.8, 0.3, 0.5, 'percent')
           >>> _= dem.plot()
           >>> b = new_dem.bounds
           >>> _= plt.plot(
           ...             [b.left, b.right, b.right, b.left, b.left],
           ...             [b.top, b.top, b.bottom, b.bottom, b.top],'r-', lw=2
           ... )
           >>> plt.show()
        """
        height, width = self.shape[0], self.shape[1]
        left_bound, right_bound = self.extent[0], self.extent[1]
        top_bound, bottom_bound = self.extent[3], self.extent[2]

        if mode == 'percent':
            if not all(0.0 <= float(val) <= 1.0
                       for val in [top, bottom, left, right]):
                err = ("All values must be between 0.0 and 1.0 for mode "
                       "'percent'.")
                raise ValueError(err) from None

            y_start = int(top * height)
            y_end = int(bottom * height)
            x_start = int(left * width)
            x_end = int(right * width)

        elif mode == 'coordinate':
            if not (all(left_bound<= val <= right_bound for
                    val in [left, right]) and
                    all(bottom_bound <= val <= top_bound for
                    val in [top, bottom])):
                err = (f"All values must be within the grid bounds: "
                      f"left: {left_bound}, right: {right_bound}, "
                      f"top: {top_bound}, bottom: {bottom_bound}.")
                raise ValueError(err) from None

            y_start = int((top_bound - top) / self.cellsize)
            y_end = int((top_bound - bottom) / self.cellsize)
            x_start = int((left - left_bound) / self.cellsize)
            x_end = int((right - left_bound) / self.cellsize)

        elif mode == 'pixel':
            if not (all(0 <= val < height for val in [top, bottom]) and
                    all(0 <= val < width for val in [left, right])):
                err = (f"All values must be within the grid pixel indices: "
                       f"left: 0-{width-1}, right: 0-{width-1}, "
                       f"top: 0-{height-1}, bottom: 0-{height-1}.")
                raise ValueError(err) from None
            y_start, y_end = int(top), int(bottom)
            x_start, x_end = int(left), int(right)

        else:
            err = (f"Invalid mode {mode} provided. Please use one of the "
                   "following modes:\n 1. 'percent' for relative values "
                   "(0.0 to 1.0)\n 2. 'coordinate' for absolute coordinates "
                   "within the grid bounds\n"
                   "3. 'pixel' for pixel indices (0 to grid width/height-1)")
            raise ValueError(err) from None

        # Ensure x_start < x_end and y_start < y_end to handle switched
        # bounds instead of raising an error
        if x_start > x_end:
            x_start, x_end = x_end, x_start
        if y_start > y_end:
            y_start, y_end = y_end, y_start

        result = cp.copy(self)

        # Calculate new transform
        new_x_origin, new_y_origin = self.transform * (x_start, y_start)
        #new_transform = self.transform * (x_start, y_start)
        new_transform = Affine(
            self.transform.a, self.transform.b, new_x_origin,
            self.transform.d, self.transform.e, new_y_origin)
        result.transform = new_transform

        # Calculate new bounds
        xs = np.array([x_start, x_end, x_end, x_start])
        ys = np.array([y_start, y_start, y_end, y_end])

        ws, zs = self.transform * (xs, ys)
        new_left = min(ws)
        new_right = max(ws)
        new_bottom = min(zs)
        new_top = max(zs)

        new_bounds = BoundingBox(new_left, new_bottom, new_right, new_top)
        result.bounds = new_bounds

        # Crop DEM
        result.z = np.copy(result.z[y_start:y_end, x_start:x_end])
        return result

    # 'Magic' functions:
    # ------------------------------------------------------------------------

    def __eq__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        dem.z = self.z == other.z

        return dem

    def __ne__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        dem.z = self.z != other.z

        return dem

    def __gt__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        dem.z = self.z > other.z

        return dem

    def __lt__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        dem.z = self.z < other.z

        return dem

    def __ge__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        dem.z = self.z >= other.z

        return dem

    def __le__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        dem.z = self.z <= other.z

        return dem

    def __add__(self, other):
        dem = cp.copy(self)

        if isinstance(other, self.__class__):
            dem.z = self.z + other.z
            return dem

        dem.z = self.z + other
        return dem

    def __sub__(self, other):
        dem = cp.copy(self)

        if isinstance(other, self.__class__):
            dem.z = self.z - other.z
            return dem

        dem.z = self.z - other
        return dem

    def __mul__(self, other):
        dem = cp.copy(self)

        if isinstance(other, self.__class__):
            dem.z = self.z * other.z
            return dem

        dem.z = self.z * other
        return dem

    def __div__(self, other):
        dem = cp.copy(self)

        if isinstance(other, self.__class__):
            dem.z = self.z / other.z
            return dem

        dem.z = self.z / other
        return dem

    def __and__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        # Check for invalid values
        if np.any((self.z != 0) & (self.z != 1)) or np.any((other.z != 0) & (other.z != 1)):
            error = "Invalid cell value. 'and' can only compare True (1) and False (0) values."
            raise ValueError(error)

        # Perform element-wise bitwise AND operation
        dem.z = np.logical_and(self.z, other.z)

        return dem

    def __or__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        # Check for invalid values
        if np.any((self.z != 0) & (self.z != 1)) or np.any((other.z != 0) & (other.z != 1)):
            error = "Invalid cell value. 'and' can only compare True (1) and False (0) values."
            raise ValueError(error)

        # Perform element-wise bitwise OR operation
        dem.z = np.logical_or(self.z, other.z)

        return dem

    def __xor__(self, other):
        dem = cp.deepcopy(self)

        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare two GridObjects.")

        if self.columns != other.columns or self.rows != other.rows:
            raise ValueError("Both GridObjects have to be the same size.")

        # Check for invalid values
        if np.any((self.z != 0) & (self.z != 1)) or np.any((other.z != 0) & (other.z != 1)):
            error = "Invalid cell value. 'and' can only compare True (1) and False (0) values."
            raise ValueError(error)

        # Perform element-wise bitwise XOR operation
        dem.z = np.logical_xor(self.z, other.z)

        return dem

    def __len__(self):
        return len(self.z)

    def __iter__(self):
        return iter(self.z)

    def __getitem__(self, index):
        return self.z[index]

    def __setitem__(self, index, value):
        try:
            value = np.float32(value)
        except (ValueError, TypeError):
            raise TypeError(
                f"{value} can't be converted to float32.") from None

        self.z[index] = value

    # pylint: disable=unused-argument
    def __array__(self, dtype=None, copy=None):
        if copy:
            return self.z.copy()
        return self.z

    def __str__(self):
        return str(self.z)

    def __repr__(self):

        # Determine the coordinate system
        str_coord = ''
        if self.georef is not None and self.georef.is_projected:
            str_coord = f'coordinate system (Projected): {self.georef}'
        elif self.georef is not None and self.georef.is_geographic:
            str_coord = f'coordinate system (Geographic): {self.georef}'
        else:
            str_coord = f'coordinate system: {self.georef}'

        return f"""name: {self.name}
        path: {self.path}
        rows: {self.rows}
        cols: {self.columns}
        cellsize: {self.cellsize}
        bounds: {self.bounds}
        transform: {self.transform}
        {str_coord}
        maximum z-value: {np.nanmax(self.z)}
        minimum z-value: {np.nanmin(self.z)}
        """



'''
Set of tests for graphflood
'''

import numpy as np
import matplotlib.pyplot as plt # TO REMOVE I USE FOR CHECKS
import topotoolbox as ttb
import pytest
from rasterio import Affine
from rasterio.coords import BoundingBox


def debug_array_info(array, name):
	print(f"{name} - ID: {id(array)}, Shape: {array.shape}, Contiguous: {array.flags['C_CONTIGUOUS'] or array.flags['F_CONTIGUOUS']}")




def _generate_south_slopping_GridObj(nrows, ncolumns, cellsize, s0, dtype:'float | np.float32 | np.float64' = np.float32):
	'''
	Generates a grid of South Slopping topography ready for graphflood.
	N/E/W are walls and South lets flow escape.
	REturns the bc array and the topo array

	Parameters
	----------
	nrows: Number of rows
	ncolumns: Number of columns
	cellsize: spatial step in m
	s0: slope
	dtype: the numpy type for the object ()


	Returns
	-------
	GridObject
		A grid object with the computed topography.
	Numpy 2D array
		A numpy array of gridobj shape with graphflood-ready boundary conditions.

	'''

	# Create a 2D array of zeros with the specified number of rows and columns
	Z = np.zeros((nrows, ncolumns), dtype = dtype)

	# Calculate the total length of the grid in the north-south direction
	total_length = nrows * cellsize

	# Create an array representing the distance from the southern boundary for each row
	distances = np.arange(nrows, dtype = np.float32) * cellsize

	# Broadcast it in 2D
	distances = np.tile(distances[:, np.newaxis], ncolumns)

	# Calculate the elevation values for each row
	Z = s0 * (total_length - distances)

	# add walls
	Z[0,:] = 9999
	Z[:,[0,-1]] = 9999

	# bcs array
	bcs = np.ones_like(Z,np.uint8) # data inside
	bcs[Z == 9999] = 0 # no data on walls
	bcs[0,:] = 0 # can out in the South
	bcs[-1,:] = 3 # can out in the South

	# Create the GridObject
	grid = ttb.GridObject()
	#
	grid.path = ''
	# name of DEM
	grid.name = 'test'

	# raster metadata
	grid.z = Z.astype(np.float32)
	grid.cellsize = cellsize  # in meters if crs.is_projected == True


	# georeference
	grid.bounds = BoundingBox(0., nrows*cellsize, ncolumns*cellsize, 0.)
	grid.transform = Affine.identity()
	grid.crs = None

	return grid, bcs



def test_analytical_solution():
	'''
	tests an analytical solution of graphflood on a small rectangular channel
	'''
	# Dimensions
	nrows, ncolumns = 126, 20 
	cellsize = 2.
	s0 = 1e-2
	dt = 1e-3

	# param hydro
	qvolin_total = 15
	manning = 0.033

	# Sloping grid
	grid, bcs = _generate_south_slopping_GridObj(nrows, ncolumns, cellsize, s0)
	# Precipitation grid
	p = np.zeros_like(grid.z)
	qvolin_unit = qvolin_total/(ncolumns-2)
	p[1,1:-1] = qvolin_unit/cellsize**2
	# Analytical solution to manning
	hwstar = (manning * qvolin_unit/(cellsize*(s0**0.5)))**(3./5.)
	
	# Simulation
	hw = ttb.run_graphflood(grid, bcs=bcs, dt=dt, p=p, manning=manning, n_iterations=5000)['hw']
	
	# Checking if it works within an error margin
	# Note that I cherry picked these value to be representative of the right solution while minimising the iterations
	assert abs(np.mean(hw[1:50,1:-1]) - hwstar) < 0.02



def test_graphflood_object():
	'''
	Tests that the graphflood object actually runs smoothly.
	Does not test the physics, tests the object/functions.
	'''

	dem = ttb.load_dem('greenriver') # example dem

	# default test
	gf = ttb.GFObject(dem)
	gf.run_n_iterations(n_iterations=1)

	# Testing with 2D numpy inputs
	# Boundary conditions (if not given to model, will default to 4 edges can out)
	# 1 = data, 3 = can out, 0 = no data, here we open every border and nodata region
	bcs = np.full_like(dem.z,1).astype(np.uint8)
	bcs[~np.isfinite(dem.z)] = 3 # no data is np.nan, so here we say where nodata can out (flow won't cross)
	bcs[[-1,0],:] = 3
	bcs[:,[-1,0]] = 3
	prec = np.full_like(dem.z,10e-3/3600) # can be 2D
	manning = np.full_like(dem.z,0.033) # can be 2D
	gf = ttb.GFObject(dem,p=prec,bcs=bcs,manning=manning)
	gf.run_n_iterations(n_iterations=1,dt=5e-3)

	# Testing with GridObject inputs
	prec = dem.duplicate_with_new_data(prec)
	bcs = dem.duplicate_with_new_data(bcs)
	manning = dem.duplicate_with_new_data(manning)
	gf = ttb.GFObject(dem,p=prec,bcs=bcs,manning=manning)
	gf.run_n_iterations(n_iterations=1,dt=5e-3)

	# getting the outputs
	tval = gf.get_qvol_i()
	tval = gf.get_qvol_o()
	tval = gf.get_q()
	tval = gf.get_u()
	tval = gf.get_sw()
	tval = gf.compute_tau()
	tval = gf.get_convergence_metrics()



"""Functions for working with flow networks

These functions often apply to both StreamObjects and FlowObjects.
"""
import numpy as np

# pylint: disable=no-name-in-module
from . import _stream  # type: ignore

def imposemin(s, dem, minimum_slope=0.0):
    """Minima imposition along a drainage network

    Parameters
    ----------
    s : StreamObject | FlowObject
        The drainage network used to determine flow directions

    dem : GridObject | np.ndarray
        The elevations to be carved. If s is a FlowObject, this should
        either be a GridObject or a 2D array of the appropriate
        shape. If s is a StreamObject, this should be a GridObject or
        a 2D array of the shape of the DEM from which the StreamObject
        was derived or a 1D array (a node attribute list) with as many
        entries as there are nodes in the stream network. The element
        type of the array can be any numerical type, but it will be
        converted into float32, so some precision may be lost.

    minimum_slope : float, optional
        The minimum downward gradient (expressed as a positive
        nondimensional slope) to be imposed on the
        elevations. Defaults to zero. Set to a small positive number
        (e.g. 0.001) to impose a shallow downward slope on the
        resulting stream profile. Too high a minimum gradient will
        result in channels that lie well below the land surface.

    Returns
    -------
    GridObject | np.ndarray
        The elevations with the minimum downward gradient imposed. If
        `dem` is a GridObject, a GridObject is returned. Otherwise a
        single-precision floating point array of the same shape as
        `dem` is returned.

    """
    result = s.ezgetnal(dem, dtype=np.float32) # This returns a copy

    d = -s.node_to_node_distance() * minimum_slope
    _stream.traverse_down_f32_min_add(result, d, s.source, s.target)

    return result



"""This module contains the GFObject class.
"""
from copy import deepcopy

import numpy as np
import matplotlib.pyplot as plt

# pylint: disable=no-name-in-module
from . import graphflood as tgf
from .grid_object import GridObject

__all__ = ['GFObject']


class GFObject():
    """GraphFlood object for stationnary, large-scale ready hydrodynamic modeling.

    See Gailleton et al. 2024 for details (https://esurf.copernicus.org/articles/12/1295/2024/)

    Contains flow information and parameters for water routing on a DEM.
    Handles boundary conditions, precipitation, and Manning roughness coefficients.
    """

    def __init__(self,
        grid: GridObject,
        bcs: np.ndarray | GridObject | None = None,
        p: float | np.ndarray | GridObject = 10 * 1e-3 / 3600,
        manning: float | np.ndarray | GridObject = 0.033
    ):
        """Initialize flood routing object.

        Parameters
        ----------
        grid : GridObject
            Topographic grid object
        bcs : ndarray or GridObject, optional
            Boundary conditions (None for default open boundaries)
        p : float or ndarray or GridObject, optional
            Precipitation rate [m/s] (default: 10mm/h)
        manning : float or ndarray or GridObject, optional
            Manning roughness coefficient (default: 0.033)
        """
        # Deep copy grid and ensure float64 precision for elevation
        self.grid = deepcopy(grid)
        self.grid.z = self.grid.z.astype(np.float64)

        # Initialize water height GridObject
        self._hw = deepcopy(self.grid)
        self._hw.z = np.zeros_like(self.grid.z)

        # Process boundary conditions
        if bcs is None:
            # Default: open boundaries on edges, internal cells are normal flow
            self._bcs = np.ones((grid.rows, grid.columns), dtype=np.uint8)
            self._bcs[[0, -1], :] = 3  # Open boundary on top/bottom
            self._bcs[:, [0, -1]] = 3  # Open boundary on left/right
        else:
            # Validate boundary condition dimensions
            if bcs.shape != grid.shape:
                raise RuntimeError(
                    "Boundary conditions must match grid dimensions"
                )
            if isinstance(bcs, GridObject):
                self._bcs = bcs.z
            else:
                self._bcs = bcs


        # Process precipitation input
        if isinstance(p, np.ndarray):
            if p.shape != grid.shape:
                raise RuntimeError(
                    "Precipitation array must match grid dimensions"
                )
            self._precipitations = p
        elif isinstance(p, GridObject):
            if p.shape != grid.shape:
                raise RuntimeError(
                    "Precipitation GridObject must match grid dimensions"
                )
            self._precipitations = p.z
        else:
            # Uniform precipitation across grid
            self._precipitations = np.full_like(self.grid.z, p)

        # Process Manning roughness coefficient
        if isinstance(manning, np.ndarray):
            if manning.shape != grid.shape:
                raise RuntimeError(
                    "Manning coefficient array must match grid dimensions"
                )
            self._manning = manning
        elif isinstance(manning, GridObject):
            if manning.shape != grid.shape:
                raise RuntimeError(
                    "Manning coefficient GridObject must match grid dimensions"
                )
            self._manning = manning.z
        else:
            # Uniform Manning coefficient across grid
            self._manning = np.full_like(self.grid.z, manning)

        # Placeholder for the results from the model
        self.res: dict = {}

    # Water height getters and setters
    @property
    def hw(self) -> GridObject:
        """Get water height GridObject."""
        return self._hw

    @hw.setter
    def hw(self, value: np.ndarray | GridObject) -> None:
        """Set water height array."""
        if isinstance(value, GridObject):
            if value.shape != self.grid.shape:
                raise ValueError("Water height GridObject must match grid dimensions")
            self._hw = deepcopy(value)
            self._hw.z = self._hw.z.astype(np.float64)
        else:
            if value.shape != self.grid.shape:
                raise ValueError("Water height array must match grid dimensions")
            self._hw.z = value.astype(np.float64)

    # Boundary conditions getters and setters
    @property
    def bcs(self) -> np.ndarray:
        """Get boundary conditions array."""
        return self._bcs

    @bcs.setter
    def bcs(self, value: np.ndarray | GridObject) -> None:
        """Set boundary conditions array."""
        if isinstance(value, GridObject):
            if value.shape != self.grid.shape:
                raise ValueError("Boundary conditions must match grid dimensions")
            self._bcs = value.z.astype(np.uint8)
        else:
            if value.size != self.grid.z.size:
                raise ValueError("Boundary conditions must match grid size")
            self._bcs = value.astype(np.uint8)

    # Precipitation getters and setters
    @property
    def precipitations(self) -> np.ndarray:
        """Get precipitation array."""
        return self._precipitations

    @precipitations.setter
    def precipitations(self, value: float | np.ndarray | GridObject) -> None:
        """Set precipitation array."""
        if isinstance(value, np.ndarray):
            if value.size != self.grid.z.size:
                raise ValueError("Precipitation array must match grid size")
            self._precipitations = value
        elif isinstance(value, GridObject):
            if value.shape != self.grid.shape:
                raise ValueError("Precipitation GridObject must match grid dimensions")
            self._precipitations = value.z
        else:
            self._precipitations = np.full_like(self.grid.z, value)

    # Manning coefficient getters and setters
    @property
    def manning(self) -> np.ndarray:
        """Get Manning roughness coefficient array."""
        return self._manning

    @manning.setter
    def manning(self, value: float | np.ndarray | GridObject) -> None:
        """Set Manning roughness coefficient array."""
        if isinstance(value, np.ndarray):
            if value.size != self.grid.z.size:
                raise ValueError("Manning coefficient array must match grid size")
            self._manning = value
        elif isinstance(value, GridObject):
            if value.shape != self.grid.shape:
                raise ValueError("Manning coefficient GridObject must match grid dimensions")
            self._manning = value.z
        else:
            self._manning = np.full_like(self.grid.z, value)

    def run_n_iterations(self, dt: float = 1e-3, sfd: bool = False,
                         d8: bool = True, n_iterations: int = 100):
        """Run graphflood model for n iterations.

        Executes the hydrodynamic simulation using the graphflood algorithm.
        Updates water height and stores flow outputs in self.res.

        Parameters
        ----------
        dt : float, optional
            Time step in seconds (default: 1e-3)
        sfd : bool, optional
            Use single flow direction routing (default: False)
        d8 : bool, optional
            Use D8 flow routing (default: True)
        n_iterations : int, optional
            Number of simulation iterations (default: 100)
        """
        # Run the graphflood simulation with current parameters
        self.res = tgf.run_graphflood(
            self.grid,
            initial_hw=self._hw.z,
            bcs=self._bcs,
            dt=dt,
            p=self._precipitations,
            manning=self._manning,
            sfd=sfd,
            d8=d8,
            n_iterations=n_iterations)

        # Update water height from results and remove from res dict
        if isinstance(self.res['hw'], GridObject):
            self._hw = self.res['hw']
        else:
            self._hw.z = self.res['hw']
        del self.res['hw']

    def run_n_iterations_dynamic_graph(self, input_qvol: np.ndarray | GridObject,
                                       dt: float = 1e-3, d8: bool = True,
                                       n_iterations: int = 100):
        """Run graphflood dynamic graph model for n iterations.

        Executes the hydrodynamic simulation using the dynamic induced graph
        graphflood algorithm. Updates water height and stores flow outputs in self.res.

        Parameters
        ----------
        input_qvol : np.ndarray or GridObject
            Input discharge locations array [m³/s].
            Must have the same shape as the grid.
        dt : float, optional
            Time step in seconds (default: 1e-3)
        d8 : bool, optional
            Use D8 flow routing (default: True)
        n_iterations : int, optional
            Number of simulation iterations (default: 100)
        """
        # Run the dynamic graph graphflood simulation with current parameters
        self.res = tgf.run_graphflood_dynamic_graph(
            self.grid,
            input_qvol=input_qvol,
            initial_hw=self._hw.z,
            bcs=self._bcs,
            dt=dt,
            p=self._precipitations,
            manning=self._manning,
            d8=d8,
            n_iterations=n_iterations)

        # Update water height from results and remove from res dict
        if isinstance(self.res['hw'], GridObject):
            self._hw = self.res['hw']
        else:
            self._hw.z = self.res['hw']
        del self.res['hw']

    def compute_input_qvol_from_area_threshold(self, area_threshold: float,
                                             d8: bool = True, step: float = 1e-3) -> GridObject:
        """Compute input discharge array from drainage area threshold.

        Identifies channel heads where drainage area crosses a threshold and assigns
        precipitation-weighted discharge at those entry points for use with the dynamic
        graph graphflood algorithm.

        Parameters
        ----------
        area_threshold : float
            Drainage area threshold for entry points [m²]
        d8 : bool, optional
            Use D8 flow routing (default: True)
        step : float, optional
            Delta_Z to apply minimum elevation increase and avoid flats during
            priority flooding (default: 1e-3)

        Returns
        -------
        GridObject
            Grid object with computed input discharge array [m³/s]
        """
        return tgf.compute_input_qvol_from_area_threshold(
            self.grid,
            area_threshold=area_threshold,
            hw=self._hw.z,
            bcs=self._bcs,
            p=self._precipitations,
            d8=d8,
            step=step)

    # Model output getters (read-only)
    @property
    def qvol_i(self) -> GridObject:
        """Incoming discharge GridObject [m³/s]

        Returns
        -------
        GridObject
            Incoming discharge for each cell

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before accessing results")
        return self.res['Qi']

    @property
    def qvol_o(self) -> GridObject:
        """Outgoing discharge GridObject [m³/s]

        Returns
        -------
        GridObject
            Outgoing discharge for each cell

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before accessing results")
        return self.res['Qo']

    @property
    def q(self) -> GridObject:
        """Specific outgoing discharge GridObject [m²/s]

        Returns
        -------
        GridObject
            Outgoing discharge per unit width for each cell

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before accessing results")
        return self.res['qo']

    @property
    def u(self) -> GridObject:
        """Flow velocity GridObject [m/s]

        Returns
        -------
        GridObject
            Flow velocity for each cell

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before accessing results")
        return self.res['u']

    @property
    def sw(self) -> GridObject:
        """Wetted area GridObject [m²]

        Returns
        -------
        GridObject
            Wetted surface area for each cell

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before accessing results")
        return self.res['Sw']

    # API methods (equivalent to property getters)
    def get_qvol_i(self) -> GridObject:
        """Incoming discharge GridObject [m³/s]
        """
        return self.qvol_i

    def get_qvol_o(self) -> GridObject:
        """Outgoing discharge GridObject [m³/s]
        """
        return self.qvol_o

    def get_q(self) -> GridObject:
        """Specific outgoing discharge GridObject [m²/s]
        """
        return self.q

    def get_u(self) -> GridObject:
        """Flow velocity GridObject [m/s]
        """
        return self.u

    def get_sw(self) -> GridObject:
        """Wetted area GridObject [m²]
        """
        return self.sw

    def compute_tau(self, gravity: float = 9.81, flow_density: float = 1000) -> GridObject:
        """Compute 2D maps of river shear stress (following Manning's approximation).

        Parameters
        ----------
        gravity : float, optional
            Gravitational acceleration [m/s²] (default: 9.81)
        flow_density : float, optional
            Water density [kg/m³] (default: 1000)

        Returns
        -------
        GridObject
            GridObject containing shear stress values [Pa]

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before computing shear stress")

        # Create mask for valid flow cells
        mask = (self._bcs > 0) & (self._hw.z > 0) & (self.sw.z > 0)

        # River shear stress calculation
        tau = np.zeros_like(self._hw.z)
        tau[mask] = self.sw.z[mask] * self._hw.z[mask] * flow_density * gravity

        # Return as GridObject with same georeferencing as input grid
        tau_grid = deepcopy(self.grid)
        tau_grid.z = tau
        return tau_grid


    def get_convergence_metrics(self):
        """Compute convergence metrics for flow conservation analysis.

        Calculates the ratio of outgoing to incoming discharge for each cell
        to assess flow conservation and model convergence quality. Data is binned
        by log10(incoming discharge) and comprehensive percentile statistics are
        computed for each bin.

        Returns
        -------
        tuple
            - convergence_map_grid : GridObject
                GridObject containing convergence ratios (Qo/Qi) for each cell
            - conv : dict
                Dictionary containing:
                - 'bin_centers': discharge bin centers [m³/s]
                - 'bin_counts': number of cells in each bin
                - 'percentile_X': percentile X of convergence ratios for each bin
                  (where X = 1, 5, 9, 13, ..., 97 - every 4th percentile)

        Raises
        ------
        RuntimeError
            If model hasn't been run yet
        """
        if not self.res:
            raise RuntimeError("Model must be run before accessing convergence metrics")

        # Create mask for valid flow cells (excluding boundaries)
        mask = self._bcs > 0

        # Initialize convergence map with default value of 1 (perfect conservation)
        convergence_map = np.zeros_like(self._hw.z)
        convergence_map[mask] = 1.0

        # Update mask to include cells with positive incoming discharge
        mask = mask & (self.qvol_i.z >= 0)

        # Calculate convergence ratio (outgoing/incoming discharge)
        # Values close to 1 indicate good conservation
        with np.errstate(divide='ignore', invalid='ignore'):
            convergence_map[mask] = self.qvol_o.z[mask] / self.qvol_i.z[mask]

        # Return as GridObject with same georeferencing as input grid
        convergence_map_grid = deepcopy(self.grid)
        convergence_map_grid.z = convergence_map

        # Calculate binned percentile statistics based on log10(qvol_i)
        conv = {}

        # Flatten arrays for binning
        qvol_i_flat = self.qvol_i.z.ravel()
        convergence_flat = convergence_map.ravel()

        # Create mask for valid data (non-NaN and positive discharge)
        valid_mask = ~np.isnan(qvol_i_flat) & ~np.isnan(convergence_flat) & (qvol_i_flat > 0)
        qvol_i_valid = qvol_i_flat[valid_mask]
        convergence_valid = convergence_flat[valid_mask]

        if len(qvol_i_valid) > 0:
            # Transform to log10 space for binning
            log_qvol_i_valid = np.log10(qvol_i_valid)

            # Create linear bins in log10 space
            n_bins = 20
            log_min = np.min(log_qvol_i_valid)
            log_max = np.max(log_qvol_i_valid)
            bin_edges = np.linspace(log_min, log_max, n_bins + 1)

            # Digitize log10(qvol_i) values into bins
            bin_indices = np.digitize(log_qvol_i_valid, bin_edges)

            # Define percentile values to calculate (every 4th from 1 to 100)
            percentile_values = np.arange(1, 101, 4)

            # Initialize storage for results
            bin_centers = []
            percentiles_dict = {f'percentile_{p}': [] for p in percentile_values}
            bin_counts = []

            # Calculate statistics for each bin
            for i in range(1, len(bin_edges)):
                bin_mask = bin_indices == i
                if np.sum(bin_mask) > 0:
                    bin_convergence = convergence_valid[bin_mask]
                    bin_log_qvol = log_qvol_i_valid[bin_mask]

                    # Store bin center in original (non-log) space
                    bin_centers.append(10**np.mean(bin_log_qvol))

                    # Calculate all percentiles for this bin
                    for p in percentile_values:
                        percentiles_dict[f'percentile_{p}'].append(
                            np.percentile(bin_convergence, p))

                    bin_counts.append(len(bin_convergence))

            # Convert to numpy arrays
            conv['bin_centers'] = np.array(bin_centers)
            conv['bin_counts'] = np.array(bin_counts)

            # Add all percentiles to conv dictionary
            for p in percentile_values:
                conv[f'percentile_{p}'] = np.array(percentiles_dict[f'percentile_{p}'])

        else:
            # No valid data - initialize empty arrays
            conv['bin_centers'] = np.array([])
            conv['bin_counts'] = np.array([])

            # Initialize empty arrays for all percentiles
            percentile_values = np.arange(1, 101, 4)
            for p in percentile_values:
                conv[f'percentile_{p}'] = np.array([])
        return convergence_map_grid, conv


    def plot_convergence(self):
        """Plot convergence metrics as spatial map and log10(discharge)-binned statistics.

        Creates a two-panel plot showing:
        1. Spatial map of convergence ratios (Qo/Qi)
        2. Convergence percentiles vs log10(incoming discharge) with uncertainty bands

        The convergence analysis bins data by log10(qvol_i) and calculates every 4th
        percentile from 1 to 100, providing comprehensive statistical information
        about flow conservation across different discharge magnitudes.

        Returns
        -------
        tuple
            - fig : matplotlib.figure.Figure
                The figure object
            - ax : array of matplotlib.axes.Axes
                Array containing the two subplot axes [spatial_map, percentiles_plot]
        """
        convergence_map_grid, conv = self.get_convergence_metrics()

        fig,ax = plt.subplots(1,2)

        self.grid.plot_hs(ax=ax[0])

        # Getting the extent
        im = ax[0].images[0]        # first image in this Axes
        ext = im.get_extent()

        im = ax[0].imshow(convergence_map_grid.z, cmap='RdBu_r',
                         vmin=0.8, vmax=1.2, extent=ext)
        plt.colorbar(im, label = 'Convergence (perfect = 1.)')


        # Plot convergence percentiles vs discharge bins
        if len(conv['bin_centers']) > 0:
            # Plot key percentiles with fill_between for uncertainty bands
            ax[1].fill_between(conv['bin_centers'], conv['percentile_5'],
                             conv['percentile_97'], alpha=0.2, color='lightblue',
                             label='5th-97th percentile')
            ax[1].fill_between(conv['bin_centers'], conv['percentile_25'],
                             conv['percentile_73'], alpha=0.3, color='blue',
                             label='25th-73rd percentile')
            ax[1].plot(conv['bin_centers'], conv['percentile_49'],
                      color='red', linewidth=2, label='~Median (49th percentile)')
            ax[1].axhline(y=1.0, color='black', linestyle='--', alpha=0.7,
                         label='Perfect conservation')
            ax[1].set_xscale('log')
            ax[1].legend()

        ax[1].grid(True, alpha=0.3)
        ax[1].set_xlabel('Incoming Discharge [m³/s]')
        ax[1].set_ylabel('Convergence Ratio (Qo/Qi)')

        return fig, ax



"""
Basic interface to libtopotoolbox implementation of graphflood.
"""

from copy import deepcopy

import numpy as np

# pylint: disable=no-name-in-module
from . import _graphflood  # type:ignore
from .grid_object import GridObject

# exposing function as string dictionary
funcdict = {
    "run_full": _graphflood.graphflood_run_full,
    "run_metrics": _graphflood.graphflood_metrics,
    "sfgraph": _graphflood.graphflood_sfgraph,
    "priority_flood_TO":
    _graphflood.compute_priority_flood_plus_topological_ordering,
    "priority_flood": _graphflood.compute_priority_flood,
    "drainage_area_single_flow": _graphflood.compute_drainage_area_single_flow,
    "dynamic_graph": _graphflood.graphflood_dynamic_graph,
    "compute_input_qvol_from_area_threshold": _graphflood.compute_input_qvol_from_area_threshold,
}

__all__ = [
    "run_graphflood",
    "run_graphflood_dynamic_graph",
    "compute_input_qvol_from_area_threshold"
    ]

def run_graphflood(
    grid: GridObject,
    initial_hw: np.ndarray | GridObject | None = None,
    bcs: np.ndarray | GridObject | None = None,
    dt: float = 1e-3,
    p: float | np.ndarray | GridObject = 10 * 1e-3 / 3600,
    manning: float | np.ndarray | GridObject = 0.033,
    sfd: bool = False,
    d8: bool = True,
    n_iterations: int = 100,
):
    """
    Runs the full graphflood's algorithm as described in Gailleton et al., 2024

    Parameters
    ----------
    grid : GridObject
        A GridObject representing the digital elevation model.
    initial_hw : np.ndarray or GridObject, optional
        Flow depth.
        Default is a matrix filled with zeros and the same shape as 'grid'.
    BCs : np.ndarray or GridObject, optional
        Boundary codes.
        Default is a matrix filled with ones except for the outermost edges,
        where values are set to 3, has same shape as 'grid'
    dt : float, optional
        time step(s ~ although this is not simulated time as we make the
        steady low assumption). Default is 1e-3.
    P : float, np.ndarray, or GridObject, optional
        Precipitation rates in m.s-1
        Default is a matrix with the same shape of 'grid'
        filled with 10 * 1e-3 / 3600.
    manning : float, np.ndarray, or GridObject, optional
        Friction coefficient.
        Default is a matrix with the same shape as 'grid' filled with 0.033.
    SFD : bool, optional
        True to compute single flow directions, False to compute multiple flow
        directions. Default is `False`.
    D8 : bool, optional
        True to include diagonal paths. Default is `True`.
    N_iterations : int, optional
        Number of iterations for the simulation. Default is 100.

    Returns
    -------
    GridObject
        A grid object with the computed water depths.

    Raises
    ------
    RuntimeError
        If the shape of `initial_hw`, `BCs`, `P`, or `manning` does not match
        the shape of the 'grid' GridObject`.

    Example
    -------
    .. plot::

       >>> import topotoolbox
       >>> import matplotlib.pyplot as plt
       >>> import matplotlib.colors as colors
       >>> flood = topotoolbox.run_graphflood(topotoolbox.load_dem('perfectworld'))
       >>> im = flood['hw'].plot(norm=colors.SymLogNorm(linthresh=1e-3, linscale=1))
       >>> plt.show()
    """

    # Preparing the arguments
    ny = grid.rows
    nx = grid.columns
    dx = grid.cellsize
    dim = np.array([ny, nx], dtype=np.uint64)

    # Order C for vectorised topography
    z = grid.z.ravel(order="C")

    # Ingesting the flow depth
    if initial_hw is None:
        hw = np.zeros_like(z)
    else:
        if initial_hw.shape != grid.z.shape:
            raise RuntimeError(
                """Feeding the model with initial flow depth
                requires a 2D numpy array or a GridObject of
                the same dimension of the topographic grid"""
            )

        if isinstance(initial_hw, GridObject):
            hw = initial_hw.z.ravel(order="C")
        else:
            hw = initial_hw.ravel(order="C")

    # Ingesting boundary condition
    if bcs is None:
        tbcs = np.ones((grid.rows, grid.columns), dtype=np.uint8)
        tbcs[[0, -1], :] = 3
        tbcs[:, [0, -1]] = 3
        tbcs = tbcs.ravel(order="C")
    else:
        if bcs.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with boundary conditions requires
                a 2D numpy array or a GridObject of the same dimension
                 of the topographic grid"""
            )

        if isinstance(bcs, GridObject):
            tbcs = bcs.z.ravel(order="C").astype(np.uint8)
        else:
            tbcs = bcs.ravel(order="C").astype(np.uint8)

    # Ingesting Precipitations
    if isinstance(p, np.ndarray):
        if p.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with precipitations requires a
                2D numpy array or a GridObject of the same
                dimension of the topographic grid"""
            )
        precipitations = p.ravel(order="C")
    elif isinstance(p, GridObject):
        if p.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with precipitations requires a
                2D numpy array or a GridObject of the same dimension
                of the topographic grid"""
            )
        precipitations = p.z.ravel(order="C")
    else:
        # in case precipitation is a scalar
        precipitations = np.full_like(z, p)

    # Ingesting manning
    if isinstance(manning, np.ndarray):
        if manning.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with manning requires a
                2D numpy array or a GridObject of the same dimension
                 of the topographic grid"""
            )
        manning = manning.ravel(order="C")
    elif isinstance(manning, GridObject):
        if manning.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with manning requires a
                2D numpy array or a GridObject of the same dimension
                of the topographic grid"""
            )
        manning = manning.z.ravel(order="C")
    else:
        # in case precipitation is a scalar
        manning = np.full_like(z, manning)

    tz = z.astype(np.float64)
    thw = hw.astype(np.float64)
    precipitations = precipitations.astype(np.float64)
    manning = manning.astype(np.float64)

    _graphflood.graphflood_run_full(
        tz, thw, tbcs, precipitations, manning,
        dim, dt, dx, sfd, d8, n_iterations, 1e-3
    )


    qvol_i = np.zeros_like(thw)
    qvol_o = np.zeros_like(thw)
    qo = np.zeros_like(thw)
    u = np.zeros_like(thw)
    sw = np.zeros_like(thw)

    _graphflood.graphflood_metrics(
        tz, thw, tbcs, precipitations, manning,
        qvol_i, qvol_o, qo, u, sw, dim, dx, d8, 1e-3 )


    res = {}

    resw = deepcopy(grid)
    resw.z = thw.reshape(grid.shape) #.astype(np.float32)
    res['hw'] = deepcopy(resw)
    resw.z = qvol_i.reshape(grid.shape) #.astype(np.float32)
    res['Qi'] = deepcopy(resw)
    resw.z = qvol_o.reshape(grid.shape) #.astype(np.float32)
    res['Qo'] = deepcopy(resw)
    resw.z = qo.reshape(grid.shape) #.astype(np.float32)
    res['qo'] = deepcopy(resw)
    resw.z = u.reshape(grid.shape) #.astype(np.float32)
    res['u'] = deepcopy(resw)
    resw.z = sw.reshape(grid.shape) #.astype(np.float32)
    res['Sw'] = deepcopy(resw)

    return res


def run_graphflood_dynamic_graph(
    grid: GridObject,
    input_qvol: np.ndarray | GridObject,
    initial_hw: np.ndarray | GridObject | None = None,
    bcs: np.ndarray | GridObject | None = None,
    dt: float = 1e-3,
    p: float | np.ndarray | GridObject = 10 * 1e-3 / 3600,
    manning: float | np.ndarray | GridObject = 0.033,
    d8: bool = True,
    n_iterations: int = 100,
):
    """
    Runs the dynamic induced graph graphflood algorithm

    Parameters
    ----------
    grid : GridObject
        A GridObject representing the digital elevation model.
    input_qvol : np.ndarray or GridObject
        Input discharge locations array [m³/s].
        Must have the same shape as 'grid'.
    initial_hw : np.ndarray or GridObject, optional
        Flow depth.
        Default is a matrix filled with zeros and the same shape as 'grid'.
    bcs : np.ndarray or GridObject, optional
        Boundary codes.
        Default is a matrix filled with ones except for the outermost edges,
        where values are set to 3, has same shape as 'grid'
    dt : float, optional
        time step (s). Default is 1e-3.
    p : float, np.ndarray, or GridObject, optional
        Precipitation rates in m.s-1
        Default is a matrix with the same shape of 'grid'
        filled with 10 * 1e-3 / 3600.
    manning : float, np.ndarray, or GridObject, optional
        Friction coefficient.
        Default is a matrix with the same shape as 'grid' filled with 0.033.
    d8 : bool, optional
        True to include diagonal paths. Default is `True`.
    n_iterations : int, optional
        Number of iterations for the simulation. Default is 100.

    Returns
    -------
    dict
        A dictionary containing GridObjects with the computed metrics:
        - 'hw': water depths
        - 'Qi': input discharge
        - 'Qo': output discharge
        - 'qo': unit discharge
        - 'u': flow velocity
        - 'Sw': wetted surface

    Raises
    ------
    RuntimeError
        If the shape of `input_qvol`, `initial_hw`, `bcs`, `p`, or `manning` does not match
        the shape of the 'grid' GridObject`.
    """

    # Preparing the arguments
    ny = grid.rows
    nx = grid.columns
    dx = grid.cellsize
    dim = np.array([ny, nx], dtype=np.uint64)

    # Order C for vectorised topography
    z = grid.z.ravel(order="C")

    # Ingesting input_qvol
    if input_qvol.shape != grid.z.shape:
        raise RuntimeError(
            """Feeding the model with input discharge requires a
            2D numpy array or a GridObject of the same dimension
            of the topographic grid"""
        )

    if isinstance(input_qvol, GridObject):
        tinput_qvol = input_qvol.z.ravel(order="C")
    else:
        tinput_qvol = input_qvol.ravel(order="C")

    # Ingesting the flow depth
    if initial_hw is None:
        hw = np.zeros_like(z)
    else:
        if initial_hw.shape != grid.z.shape:
            raise RuntimeError(
                """Feeding the model with initial flow depth
                requires a 2D numpy array or a GridObject of
                the same dimension of the topographic grid"""
            )

        if isinstance(initial_hw, GridObject):
            hw = initial_hw.z.ravel(order="C")
        else:
            hw = initial_hw.ravel(order="C")

    # Ingesting boundary condition
    if bcs is None:
        tbcs = np.ones((grid.rows, grid.columns), dtype=np.uint8)
        tbcs[[0, -1], :] = 3
        tbcs[:, [0, -1]] = 3
        tbcs = tbcs.ravel(order="C")
    else:
        if bcs.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with boundary conditions requires
                a 2D numpy array or a GridObject of the same dimension
                 of the topographic grid"""
            )

        if isinstance(bcs, GridObject):
            tbcs = bcs.z.ravel(order="C").astype(np.uint8)
        else:
            tbcs = bcs.ravel(order="C").astype(np.uint8)

    # Ingesting Precipitations
    if isinstance(p, np.ndarray):
        if p.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with precipitations requires a
                2D numpy array or a GridObject of the same
                dimension of the topographic grid"""
            )
        precipitations = p.ravel(order="C")
    elif isinstance(p, GridObject):
        if p.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with precipitations requires a
                2D numpy array or a GridObject of the same dimension
                of the topographic grid"""
            )
        precipitations = p.z.ravel(order="C")
    else:
        # in case precipitation is a scalar
        precipitations = np.full_like(z, p)

    # Ingesting manning
    if isinstance(manning, np.ndarray):
        if manning.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with manning requires a
                2D numpy array or a GridObject of the same dimension
                 of the topographic grid"""
            )
        manning = manning.ravel(order="C")
    elif isinstance(manning, GridObject):
        if manning.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with manning requires a
                2D numpy array or a GridObject of the same dimension
                of the topographic grid"""
            )
        manning = manning.z.ravel(order="C")
    else:
        # in case precipitation is a scalar
        manning = np.full_like(z, manning)

    tz = z.astype(np.float64)
    thw = hw.astype(np.float64)
    tinput_qvol = tinput_qvol.astype(np.float64)
    precipitations = precipitations.astype(np.float64)
    manning = manning.astype(np.float64)

    qwin = np.zeros_like(thw)

    _graphflood.graphflood_dynamic_graph(
        tz, thw, tbcs, precipitations, manning, tinput_qvol, qwin,
        dim, dt, dx, d8, n_iterations
    )

    qvol_i = np.zeros_like(thw)
    qvol_o = np.zeros_like(thw)
    qo = np.zeros_like(thw)
    u = np.zeros_like(thw)
    sw = np.zeros_like(thw)

    _graphflood.graphflood_metrics(
        tz, thw, tbcs, precipitations, manning,
        qvol_i, qvol_o, qo, u, sw, dim, dx, d8, 1e-3)

    res = {}

    resw = deepcopy(grid)
    resw.z = thw.reshape(grid.shape)
    res['hw'] = deepcopy(resw)
    resw.z = qwin.reshape(grid.shape)
    res['Qi'] = deepcopy(resw)
    resw.z = qvol_o.reshape(grid.shape)
    res['Qo'] = deepcopy(resw)
    resw.z = qo.reshape(grid.shape)
    res['qo'] = deepcopy(resw)
    resw.z = u.reshape(grid.shape)
    res['u'] = deepcopy(resw)
    resw.z = sw.reshape(grid.shape)
    res['Sw'] = deepcopy(resw)

    return res


def compute_input_qvol_from_area_threshold(
    grid: GridObject,
    area_threshold: float,
    hw: np.ndarray | GridObject | None = None,
    bcs: np.ndarray | GridObject | None = None,
    p: float | np.ndarray | GridObject = 10 * 1e-3 / 3600,
    d8: bool = True,
    step: float = 1e-3,
):
    """
    Compute input discharge array for dynamic graph from drainage area threshold.
    Identifies channel heads where drainage area crosses a threshold and assigns
    precipitation-weighted discharge at those entry points.

    Parameters
    ----------
    grid : GridObject
        A GridObject representing the digital elevation model.
    area_threshold : float
        Drainage area threshold for entry points [m²].
    hw : np.ndarray or GridObject, optional
        Field of water depth [m].
        Default is a matrix filled with zeros and the same shape as 'grid'.
    bcs : np.ndarray or GridObject, optional
        Boundary codes.
        Default is a matrix filled with ones except for the outermost edges,
        where values are set to 3, has same shape as 'grid'
    p : float, np.ndarray, or GridObject, optional
        Precipitation rates in m.s-1
        Default is a matrix with the same shape of 'grid'
        filled with 10 * 1e-3 / 3600.
    d8 : bool, optional
        True to include diagonal paths. Default is `True`.
    step : float, optional
        Delta_Z to apply minimum elevation increase and avoid flats during
        priority flooding. Default is 1e-3.

    Returns
    -------
    GridObject
        A grid object with the computed input discharge array [m³/s].

    Raises
    ------
    RuntimeError
        If the shape of `hw`, `bcs`, or `p` does not match
        the shape of the 'grid' GridObject`.
    """

    # Preparing the arguments
    ny = grid.rows
    nx = grid.columns
    dx = grid.cellsize
    dim = np.array([ny, nx], dtype=np.uint64)

    # Order C for vectorised topography
    z = grid.z.ravel(order="C")

    # Initialize output array
    input_qvol = np.zeros_like(z)

    # Ingesting the water depth
    if hw is None:
        thw = np.zeros_like(z)
    else:
        if hw.shape != grid.z.shape:
            raise RuntimeError(
                """Feeding the model with water depth requires a
                2D numpy array or a GridObject of the same dimension
                of the topographic grid"""
            )

        if isinstance(hw, GridObject):
            thw = hw.z.ravel(order="C")
        else:
            thw = hw.ravel(order="C")

    # Ingesting boundary condition
    if bcs is None:
        tbcs = np.ones((grid.rows, grid.columns), dtype=np.uint8)
        tbcs[[0, -1], :] = 3
        tbcs[:, [0, -1]] = 3
        tbcs = tbcs.ravel(order="C")
    else:
        if bcs.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with boundary conditions requires
                a 2D numpy array or a GridObject of the same dimension
                 of the topographic grid"""
            )

        if isinstance(bcs, GridObject):
            tbcs = bcs.z.ravel(order="C").astype(np.uint8)
        else:
            tbcs = bcs.ravel(order="C").astype(np.uint8)

    # Ingesting Precipitations
    if isinstance(p, np.ndarray):
        if p.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with precipitations requires a
                2D numpy array or a GridObject of the same
                dimension of the topographic grid"""
            )
        precipitations = p.ravel(order="C")
    elif isinstance(p, GridObject):
        if p.shape != grid.shape:
            raise RuntimeError(
                """Feeding the model with precipitations requires a
                2D numpy array or a GridObject of the same dimension
                of the topographic grid"""
            )
        precipitations = p.z.ravel(order="C")
    else:
        # in case precipitation is a scalar
        precipitations = np.full_like(z, p)

    tz = z.astype(np.float64)
    thw = thw.astype(np.float64)
    input_qvol = input_qvol.astype(np.float64)
    precipitations = precipitations.astype(np.float64)

    _graphflood.compute_input_qvol_from_area_threshold(
        input_qvol, tz, thw, tbcs, precipitations,
        area_threshold, dim, dx, d8, step
    )

    res = deepcopy(grid)
    res.z = input_qvol.reshape(grid.shape)

    return res



import numpy as np
import pytest
from rasterio import Affine, CRS
from rasterio.coords import BoundingBox

import topotoolbox as topo

import opensimplex


@pytest.fixture
def square_dem():
    return topo.gen_random(rows=128, columns=128, seed=12)


@pytest.fixture
def wide_dem():
    return topo.gen_random(rows=64, columns=128, seed=12)


@pytest.fixture
def tall_dem():
    return topo.gen_random(rows=128, columns=64, seed=12)


@pytest.fixture
def order_dems():
    opensimplex.seed(12)

    x = np.arange(0, 128)
    y = np.arange(0, 256)

    cdem = topo.GridObject()
    cdem.z = np.array(
        64 * (opensimplex.noise2array(x/13, y/13) + 1), dtype=np.float32)
    cdem.cellsize = 13.0
    cdem.transform = Affine.permutation() * Affine.rotation(90) * Affine.scale(cdem.cellsize)
    cdem.georef = CRS.from_epsg(3857)
    top_left = cdem.transform * (x[0], y[0])
    bottom_right = cdem.transform * (x[-1], y[-1])
    cdem.bounds = BoundingBox(top=top_left[1], left = top_left[0],
                              bottom=bottom_right[1], right = bottom_right[0])

    fdem = topo.GridObject()
    fdem.z = np.asfortranarray(cdem.z)
    fdem.cellsize = 13.0
    fdem.transform = Affine.permutation() * Affine.rotation(90) * Affine.scale(fdem.cellsize)
    fdem.georef = CRS.from_epsg(3857)

    top_left = fdem.transform * (x[0], y[0])
    bottom_right = fdem.transform * (x[-1], y[-1])
    fdem.bounds = BoundingBox(top=top_left[1], left = top_left[0],
                              bottom=bottom_right[1], right = bottom_right[0])

    return [cdem, fdem]


@pytest.fixture
def types_dems():
    opensimplex.seed(12)

    x = np.arange(0, 128)
    y = np.arange(0, 256)

    dem64 = topo.GridObject()
    dem64.z = np.array(
        64 * (opensimplex.noise2array(x/13, y/13) + 1), dtype=np.float64)
    dem64.cellsize = 13.0
    dem64.transform = Affine.scale(dem64.cellsize)

    dem32 = topo.GridObject()
    dem32.z = np.array(dem64, dtype=np.float32)
    dem32.cellsize = 13.0
    dem32.transform = Affine.scale(dem32.cellsize)

    return [dem32, dem64]


def test_fillsinks(square_dem, wide_dem, tall_dem):
    # TODO: add more tests
    for grid in [square_dem, wide_dem, tall_dem]:
        # since grid is a fixture, it has to be assigned/called first
        dem = grid
        original_dem = dem.z.copy()
        filled_dem = dem.fillsinks(hybrid=False)
        filled_dem_hybrid = dem.fillsinks(hybrid=True)

        assert topo.validate_alignment(dem, filled_dem)
        assert topo.validate_alignment(dem, filled_dem_hybrid)

        # Ensure that DEM has not been modified by fillsinks
        assert np.all(original_dem == dem.z)

        # The sequential and hybrid reconstruction algorithms should
        # produce the same result
        assert np.all(filled_dem.z == filled_dem_hybrid.z)

        # Loop over all cells of the DEM
        for i in range(dem.shape[0]):
            for j in range(dem.shape[1]):

                # Test: no filled cell lower than before calling fillsinks
                assert dem[i, j] <= filled_dem[i, j]

                # Test: cell isn't a sink
                sink = 0
                for i_offset, j_offset in [
                        (-1, -1),
                        (-1, 0),
                        (-1, 1),
                        (0, -1),
                        (0, 1),
                        (1, -1),
                        (1, 0),
                        (1, 1)]:

                    i_neighbor = i + i_offset
                    j_neighbor = j + j_offset

                    if (i_neighbor < 0 or i_neighbor >= dem.z.shape[0]
                            or j_neighbor < 0 or j_neighbor >= dem.z.shape[1]):
                        continue

                    if filled_dem[i_neighbor, j_neighbor] > filled_dem[i, j]:
                        sink += 1

                assert sink < 8

@pytest.mark.parametrize("hybrid",[True, False])
@pytest.mark.parametrize("bc",[True, False])
def test_fillsinks_order(order_dems, hybrid, bc):
    cdem, fdem = order_dems

    if bc:
        bc = np.ones_like(cdem.z, dtype=np.uint8)
        bc[1:-1, 1:-1] = 0

        bc = cdem.duplicate_with_new_data(bc)
    else:
        bc = None

    cfilled = cdem.fillsinks(bc=bc, hybrid=hybrid)
    assert cfilled.z.flags.c_contiguous

    ffilled = fdem.fillsinks(bc=bc, hybrid=hybrid)
    assert ffilled.z.flags.f_contiguous

    assert np.array_equal(ffilled, cfilled)

    assert topo.validate_alignment(fdem, ffilled)
    assert topo.validate_alignment(cdem, cfilled)


def test_identifyflats(square_dem, wide_dem, tall_dem):
    # TODO: add more tests
    for dem in [square_dem, wide_dem, tall_dem]:
        sills, flats = dem.identifyflats()

        assert topo.validate_alignment(dem, sills)
        assert topo.validate_alignment(dem, flats)

        for i in range(dem.shape[0]):
            for j in range(dem.shape[1]):

                for i_offset, j_offset in [
                        (-1, -1),
                        (-1, 0),
                        (-1, 1),
                        (0, -1),
                        (0, 1),
                        (1, -1),
                        (1, 0),
                        (1, 1)]:

                    i_neighbor = i + i_offset
                    j_neighbor = j + j_offset

                    if (i_neighbor < 0 or i_neighbor >= dem.z.shape[0]
                            or j_neighbor < 0 or j_neighbor >= dem.z.shape[1]):
                        continue

                    if flats[i_neighbor, j_neighbor] < flats[i, j]:
                        assert flats[i, j] == 1.0


def test_identifyflats_order(order_dems):
    cdem, fdem = order_dems

    cdem_filled = cdem.fillsinks()
    fdem_filled = fdem.fillsinks()

    craw = cdem_filled.identifyflats(raw=True)[0]
    fraw = fdem_filled.identifyflats(raw=True)[0]

    assert np.array_equal(craw, fraw)

    assert craw.flags.c_contiguous
    assert fraw.flags.f_contiguous

    cflats, csills = cdem_filled.identifyflats()
    fflats, fsills = fdem_filled.identifyflats()

    assert np.array_equal(cflats, fflats)
    assert np.array_equal(csills, fsills)

    assert cflats.z.flags.c_contiguous
    assert csills.z.flags.c_contiguous

    assert fflats.z.flags.f_contiguous
    assert fsills.z.flags.f_contiguous


def test_excesstopography(square_dem):
    # TODO: add more tests
    with pytest.raises(TypeError):
        square_dem.excesstopography(threshold='0.1')


@pytest.fixture(name="threshold_slopes", params=["scalar", "array", "gridobject"])
def threshold_slopes(request, order_dems):
    dem = order_dems[0]

    rng = np.random.default_rng(217412861091418638741329610000239956692)
    t = rng.random(dem.shape)
    if request.param == "scalar":
        return 0.2
    elif request.param == "array":
        return t
    else:
        return dem.duplicate_with_new_data(t)


@pytest.mark.parametrize("method", ['fsm2d', 'fmm2d'])
def test_excesstopography_order(order_dems, method, threshold_slopes):
    cdem, fdem = order_dems

    cext = cdem.excesstopography(threshold=threshold_slopes, method=method)
    assert cext.z.flags.c_contiguous

    fext = fdem.excesstopography(threshold=threshold_slopes, method=method)
    assert fext.z.flags.f_contiguous

    assert np.allclose(fext, cext)

    assert topo.validate_alignment(fdem, fext)
    assert topo.validate_alignment(cdem, cext)


@pytest.mark.parametrize("fused", [True, False])
def test_hillshade_order(order_dems, fused):
    cdem, fdem = order_dems

    for azimuth in np.arange(0.0, 360.0, 2.3):
        hc = cdem.hillshade(fused=fused, azimuth=azimuth)
        hf = fdem.hillshade(fused=fused, azimuth=azimuth)
        assert np.allclose(hc, hf)

        assert topo.validate_alignment(cdem, hc)
        assert topo.validate_alignment(fdem, hf)


def test_hillshade_fused(wide_dem):
    for azimuth in np.arange(0.0, 360.0, 2.3):
        h1 = wide_dem.hillshade(azimuth=azimuth, fused=True)
        h2 = wide_dem.hillshade(azimuth=azimuth, fused=False)
        assert np.allclose(h1, h2)

        assert topo.validate_alignment(wide_dem, h1)
        assert topo.validate_alignment(wide_dem, h2)


def test_hillshade_types(types_dems):
    dem32, dem64 = types_dems

    for azimuth in np.arange(0.0, 360.0, 2.3):
        h32 = dem32.hillshade(azimuth=azimuth)
        h64 = dem64.hillshade(azimuth=azimuth)
        assert np.allclose(h64, h32)


def test_filter_order(order_dems):
    cdem, fdem = order_dems

    for method in ['mean', 'average', 'median',
                   'sobel', 'scharr', 'wiener', 'std']:
        cfiltered = cdem.filter(method=method)
        ffiltered = fdem.filter(method=method)

        assert np.array_equal(cfiltered, ffiltered)

@pytest.mark.parametrize("unit", ['tangent', 'radian', 'degree', 'sine', 'percent'])
@pytest.mark.parametrize("mp", [True, False])
def test_gradient8_order(order_dems, unit, mp):
    cdem, fdem = order_dems

    cgradient = cdem.gradient8(unit=unit, multiprocessing=mp)
    fgradient = fdem.gradient8(unit=unit, multiprocessing=mp)

    assert np.array_equal(cgradient, fgradient)

    assert cgradient.z.flags.c_contiguous
    assert fgradient.z.flags.f_contiguous


@pytest.mark.parametrize("ctype", ['profc', 'planc', 'tangc', 'meanc', 'total'])
@pytest.mark.parametrize("meanfilt", [True, False])
def test_curvature_order(order_dems, ctype, meanfilt):
    cdem, fdem = order_dems

    ccurv = cdem.curvature(meanfilt=meanfilt, ctype=ctype)
    fcurv = fdem.curvature(meanfilt=meanfilt, ctype=ctype)

    assert np.array_equal(ccurv, fcurv)


def test_dilate_order(order_dems):
    cdem, fdem = order_dems

    cdilated = cdem.dilate((3, 3))
    fdilated = fdem.dilate((3, 3))

    assert np.array_equal(cdilated, fdilated)


def test_erode_order(order_dems):
    cdem, fdem = order_dems

    ceroded = cdem.erode((3, 3))
    feroded = fdem.erode((3, 3))

    assert np.array_equal(ceroded, feroded)


@pytest.mark.parametrize("modified", [True, False])
def test_evansslope_order(order_dems, modified):
    cdem, fdem = order_dems

    cslope = cdem.evansslope(modified=modified)
    fslope = fdem.evansslope(modified=modified)

    assert np.array_equal(cslope, fslope)

    cx, cy = cdem.evansslope(modified=modified, partial_derivatives=True)
    fx, fy = fdem.evansslope(modified=modified, partial_derivatives=True)

    assert np.array_equal(cx, fx)
    assert np.array_equal(cy, fy)

def test_aspect_order(order_dems):
    cdem, fdem = order_dems

    caspect = cdem.aspect()
    faspect = fdem.aspect()

    assert np.array_equal(caspect, faspect)

    caspect_edges = cdem.aspect(classify=True)
    faspect_edges = fdem.aspect(classify=True)

    assert np.array_equal(caspect_edges, faspect_edges)

@pytest.mark.parametrize("hybrid", [True, False])
def test_prominence_order(order_dems, hybrid):
    cdem, fdem = order_dems

    cp, (cx, cy) = cdem.prominence(10.0, use_hybrid=hybrid)
    fp, (fx, fy) = fdem.prominence(10.0, use_hybrid=hybrid)

    assert np.array_equal(cp, fp)
    assert np.array_equal(cx, fx)
    assert np.array_equal(cy, fy)


def test_shufflelabel(order_dems):
    cdem, fdem = order_dems

    cfd = topo.FlowObject(cdem)
    ffd = topo.FlowObject(fdem)

    cdb = cfd.drainagebasins()
    cs = cdb.shufflelabel()
    fdb = ffd.drainagebasins()
    fs = fdb.shufflelabel()

    assert cs.shape == cdb.shape
    assert fs.shape == fdb.shape

def test_reproject_order(order_dems):
    cdem, fdem = order_dems

    c2 = cdem.reproject(CRS.from_epsg(32631))
    assert c2.z.flags.c_contiguous

    f2 = fdem.reproject(CRS.from_epsg(32631))
    assert f2.z.flags.f_contiguous

    assert np.allclose(f2, c2, equal_nan=True)

def test_resample_order(order_dems):
    cdem, fdem = order_dems

    c2 = cdem.resample(15.0)
    f2 = fdem.resample(15.0)

    assert np.allclose(f2, c2)

def test_zscore_order(order_dems):
    cdem, fdem = order_dems

    cz = cdem.zscore()
    fz = fdem.zscore()

    assert np.array_equal(cz, fz)

def test_crop_order(order_dems):
    cdem, fdem = order_dems

    c1 = cdem.crop(0.6, 0.8, 0.3, 0.5, 'percent')
    f1 = fdem.crop(0.6, 0.8, 0.3, 0.5, 'percent')

    assert np.array_equal(c1, f1)

    c2 = cdem.crop(20.0, 100.0, -20.0, -60.0, 'coordinate')
    f2 = fdem.crop(20.0, 100.0, -20.0, -60.0, 'coordinate')

    assert np.array_equal(c2, f2)

    c3 = cdem.crop(20, 50, 30, 40, 'pixel')
    f3 = fdem.crop(20, 50, 30, 40, 'pixel')

    assert np.array_equal(c3, f3)



"""Provide general utility functions for topotoolbox.
"""
import sys
import os
import random
from shutil import rmtree
from urllib.request import urlopen, urlretrieve
import urllib.error

import rasterio
import numpy as np

from .grid_object import GridObject
from .stream_object import StreamObject

__all__ = ["load_dem", "get_dem_names", "read_tif", "gen_random", "write_tif",
           "gen_random_bool", "get_cache_contents", "clear_cache",
           "read_from_cache", "load_opentopography", "write_shapefile", "get_dtype",
           "transform_coords"]



DEM_SOURCE = "https://raw.githubusercontent.com/TopoToolbox/DEMs/master"
DEM_NAMES = f"{DEM_SOURCE}/dem_names.txt"
OPEN_TOPO_SOURCE = "https://portal.opentopography.org/API/globaldem"
OPEN_TOPO_DATASETS = ('SRTMGL3', 'SRTMGL1', 'SRTMGL1_E', 'AW3D30', 'AW3D30_E',
                      'SRTM15Plus', 'NASADEM', 'COP30', 'COP90', 'EU_DTM',
                      'GEDI_L3', 'GEBCOIceTopo', 'GEBCOSubIceTopo',
                      'CA_MRDEM_DTM', 'CA_MRDEM_DSM')


def transform_coords(grid: GridObject, data1, data2=None, input_mode="indices2D",
                     output_mode="coordinates", center=True):
    """
    Helper function to convert between coordinates, indices2D, and indices1D.

    Parameters
    ----------
    grid : GridObject
        The reference grid containing transform and shape.
    data1 : array-like
        First coordinate/index component (e.g., Row indices, Node indices, or X coordinates).
    data2 : array-like, optional
        Second coordinate/index component (e.g., Col indices, or Y coordinates).
    input_mode : str
        "indices2D", "indices1D", or "coordinates".
    output_mode : str
        "indices2D", "indices1D", or "coordinates".
    center : bool, optional
        If True (default) and output_mode is "coordinates", returns coordinates at pixel centers.
        If False, returns coordinates at pixel corners.
    """
    # 1. Standardize input to indices2D (rows, cols)
    if input_mode == "indices2D":
        r, c = np.asarray(data1), np.asarray(data2)
    elif input_mode == "indices1D":
        idx = np.asarray(data1)
        if grid.z.flags.f_contiguous:
            # F-order: idx = row + col * nrows
            r = idx % grid.rows
            c = idx // grid.rows
        else:
            # C-order: idx = row * ncols + col
            r = idx // grid.columns
            c = idx % grid.columns
    elif input_mode == "coordinates":
        inv_transform = ~grid.transform
        c, r = inv_transform * (np.asarray(data1), np.asarray(data2))
    else:
        raise ValueError(f"Invalid input_mode: {input_mode}")

    # 2. Convert to requested output
    if output_mode == "indices2D":
        return r, c
    if output_mode == "indices1D":
        if grid.z.flags.f_contiguous:
            return r.astype(np.intp) + c.astype(np.intp) * grid.rows
        return r.astype(np.intp) * grid.z.shape[1] + c.astype(np.intp)
    if output_mode == "coordinates":
        if center:
            # Shift by 0.5 pixel to move from corner to center
            x, y = grid.transform * (c + 0.5, r + 0.5)
        else:
            x, y = grid.transform * (c, r)
        return np.array(x), np.array(y)

    raise ValueError(f"Invalid output_mode: {output_mode}")

def write_tif(dem: GridObject, path: str) -> None:
    """
    Write a GridObject instance to a GeoTIFF file.

    Parameters
    ----------
    dem : GridObject
        The GridObject instance to be written to a GeoTIFF file.
    path : str
        The file path where the GeoTIFF will be saved.

    Raises
    ------
    TypeError
        If `dem` is not an instance of GridObject.

    Example
    -------
    >>> import topotoolbox
    >>> dem = topotoolbox.load_dem('perfectworld')
    >>> topotoolbox.write_tif(dem, 'dem.tif')
    """

    if not isinstance(dem, GridObject):
        err = "The provided dem is not an instance of GridObject."
        raise TypeError(err) from None

    with rasterio.open(
            fp=path,
            mode='w',
            count=1,
            driver='GTiff',
            height=dem.rows,
            width=dem.columns,
            dtype=np.float32,
            crs=dem.georef,
            transform=dem.transform
    ) as dataset:
        dataset.write(dem.z, 1)
        dataset.write_mask(~np.isnan(dem.z))

def write_shapefile(streamobject: StreamObject, path: str) -> None:
    """
    Write a StreamObject instance to a shapefile. Filetype is inferred from
    the file extension. This function is based on the geopandas GeoDataFrame
    'to_file' method and supports the same formats (.shp, .gpkg, .geojson).

    Parameters
    ----------
    streamobject : StreamObject
        The StreamObject instance to be written to a shapefile.
    path : str
        The file path where the shapefile will be saved.

    Raises
    ------
    TypeError
        If `streamobject` is not an instance of StreamObject.

    Example
    -------
    >>> topotoolbox.write_shapefile(streamobject, 'streamobject.shp')

    >>> topotoolbox.write_shapefile(streamobject, 'streamobject.gpkg')

    >>> topotoolbox.write_shapefile(streamobject, 'streamobject.geojson')
    """

    if not isinstance(streamobject, StreamObject):
        err = "The provided streamobject is not an instance of StreamObject."
        raise TypeError(err) from None

    streamobject.to_shapefile(path)


def get_dtype(obj):
    """Get the dtype of a GridObject, numpy array, or scalar.

    Parameters
    ----------
    obj : GridObject | np.ndarray | scalar
        The object to get the dtype from.

    Returns
    -------
    np.dtype
        The numpy dtype of the object.

    Example
    -------
    >>> dem = topotoolbox.load_dem('taiwan')
    >>> dtype = topotoolbox.get_dtype(dem)
    >>> print(dtype)
    """
    if isinstance(obj, GridObject):
        return obj.z.dtype
    if isinstance(obj, np.ndarray):
        return obj.dtype
    return np.array(obj).dtype


def read_tif(path: str) -> GridObject:
    """Generate a new GridObject from a .tif file.

    Parameters
    ----------
    path : str
        path to .tif file.

    Returns
    -------
    GridObject
        A new GridObject of the .tif file.

    Example
    -------
    >>> import topotoolbox
    >>> dem = topotoolbox.read_tif('dem.tif') # doctest: +SKIP
    """

    grid = GridObject()

    if path is not None:
        try:
            dataset = rasterio.open(path)

        except TypeError as err:
            raise TypeError(err) from None
        except Exception as err:
            raise ValueError(err) from None

        grid.path = path
        grid.name = os.path.splitext(os.path.basename(grid.path))[0]

        grid.z = dataset.read(1, masked=True, out_dtype=np.float32).filled(np.nan)

        grid.cellsize = dataset.res[0]
        grid.bounds = dataset.bounds
        grid.transform = dataset.transform
        grid.georef = dataset.crs

    return grid


def gen_random(hillsize: int = 24, rows: int = 128, columns: int = 128,
               cellsize: float = 10.0, seed: int = 3,
               name: str = 'random grid') -> 'GridObject':
    """Generate a GridObject instance that is generated with OpenSimplex noise.

    Parameters
    ----------
    hillsize : int, optional
        Controls the "smoothness" of the generated terrain. Defaults to 24.
    rows : int, optional
        Number of rows. Defaults to 128.
    columns : int, optional
        Number of columns. Defaults to 128.
    cellsize : float, optional
        Size of each cell in the grid. Defaults to 10.0.
    seed : int, optional
        Seed for the terrain generation. Defaults to 3
    name : str, optional
        Name for the generated GridObject. Defaults to 'random grid'

    Raises
    ------
    ImportError
        If OpenSimplex has not been installed.

    Returns
    -------
    GridObject
        An instance of GridObject with randomly generated values.

    Example
    -------
    .. plot::

       >>> import topotoolbox
       >>> import matplotlib.pyplot as plt
       >>> dem = topotoolbox.gen_random(seed=8192)
       >>> _ = dem.plot(cmap='terrain')
       >>> plt.show()
    """
    try:
        import opensimplex as simplex  # pylint: disable=C0415

    except ImportError:
        err = ("For gen_random to work, use \"pip install topotool" +
               "box[opensimplex]\" or \"pip install .[opensimplex]\"")
        raise ImportError(err) from None

    noise_array = np.empty((rows, columns), dtype=np.float32)

    simplex.seed(seed)
    for y in range(0, rows):
        for x in range(0, columns):
            value = simplex.noise4(x / hillsize, y / hillsize, 0.0, 0.0)
            color = int((value + 1) * 128)
            noise_array[y, x] = color

    grid = GridObject()

    grid.z = noise_array
    grid.cellsize = cellsize
    grid.name = name
    return grid


def gen_random_bool(
        rows: int = 32, columns: int = 32, cellsize: float = 10.0,
        name: str = 'random grid') -> 'GridObject':
    """Generate a GridObject instance that contains only randomly generated
    Boolean values.

    Parameters
    ----------
    rows : int, optional
        Number of rows. Defaults to 32.
    columns : int, optional
        Number of columns. Defaults to 32.
    cellsize : float, optional
        Size of each cell in the grid. Defaults to 10.0.

    Returns
    -------
    GridObject
        An instance of GridObject with randomly generated Boolean values.

    Example
    -------
    .. plot::

       >>> import topotoolbox
       >>> import matplotlib.pyplot as plt
       >>> dem = topotoolbox.gen_random_bool()
       >>> _= dem.plot(cmap='grey')
       >>> plt.show()
    """
    bool_array = np.empty((rows, columns), dtype=np.float32)

    for y in range(0, rows):
        for x in range(0, columns):
            bool_array[x][y] = random.choice([0, 1])

    grid = GridObject()

    grid.path = ''
    grid.z = bool_array
    grid.cellsize = cellsize
    grid.name = name

    return grid


def get_dem_names() -> list[str]:
    """Returns a list of provided example Digital Elevation Models (DEMs).
    Requires internet connection to download available names.

    Returns
    -------
    list[str]
        A list of strings, where each string is the name of a DEM.

    Example
    -------
    >>> import topotoolbox
    >>> print(topotoolbox.get_dem_names()) # doctest: +SKIP
    """
    with urlopen(DEM_NAMES) as dem_names:
        dem_names = dem_names.read().decode()

    return dem_names.splitlines()


def load_dem(dem: str, cache: bool = True) -> GridObject:
    """Downloads a DEM from the TopoToolbox/DEMs repository.
    Find possible names by using 'get_dem_names()'.

    Parameters
    ----------
    dem : str
        Name of the DEM to be downloaded.
    cache : bool, optional
        If true, the DEM will be cached. Defaults to True.

    Returns
    -------
    GridObject
        A GridObject generated from the downloaded DEM.

    Example
    -------
    .. plot::

       >>> import topotoolbox
       >>> import matplotlib.pyplot as plt
       >>> dem = topotoolbox.load_dem('taiwan')
       >>> _= dem.plot(cmap='terrain')
       >>> plt.show()
    """
    if dem not in get_dem_names():
        err = ("Selected DEM has to be selected from the provided examples." +
               " See which DEMs are available by using 'get_dem_names()'.")
        raise ValueError(err)

    url = f"{DEM_SOURCE}/{dem}.tif"

    if cache:
        cache_path = os.path.join(get_save_location(), f"{dem}.tif")

        if not os.path.exists(cache_path):
            urlretrieve(url, cache_path)

        full_path = cache_path
    else:
        full_path = url

    grid_object = read_tif(full_path)

    return grid_object


def get_save_location() -> str:
    """Generates filepath to file saved in cache.

    Returns
    -------
    str
        Filepath to file saved in cache.

    Example
    -------
    >>> import topotoolbox
    >>> print(topotoolbox.utils.get_save_location()) # doctest: +ELLIPSIS
    """
    system = sys.platform

    if system == "win32":
        path = os.getenv('LOCALAPPDATA')
        if path is None:
            raise EnvironmentError(
                "LOCALAPPDATA environment variable is not set." +
                " Unable to generate path to cache.") from None
        path = os.path.join(path, "topotoolbox")

    elif system == 'darwin':
        path = os.path.expanduser('~/Library/Caches')
        path = os.path.join(path, "topotoolbox")

    else:
        path = os.path.expanduser('~/.cache')
        path = os.path.join(path, "topotoolbox")

    if not os.path.exists(path):
        os.makedirs(path)

    return path


def clear_cache(filename: str | None = None) -> None:
    """Deletes the cache directory and its contents. Can also delete a single
    file when using the argument filename. To get the contents of your cache,
    use 'get_cache_contents()'.

    Parameters
    ----------
    filename : str, optional
        Add a filename if only one specific file is to be deleted.
        Defaults to None.

    Example
    -------
    >>> import topotoolbox
    >>> topotoolbox.clear_cache()
    """
    path = get_save_location()

    if filename:
        path = os.path.join(path, filename)

    if os.path.exists(path):
        if os.path.isdir(path):
            # using shutil.rmtree since os.rmdir requires dir to be empty.
            rmtree(path)
        else:
            os.remove(path)
    else:
        print("Cache directory or file does not exist.")


def get_cache_contents() -> (list[str] | None):
    """Returns the contents of the cache directory.

    Returns
    -------
    list[str]
        List of all files in the TopoToolbox cache. If cache does
        not exist, None is returned.

    Example
    -------
    >>> import topotoolbox
    >>> print(topotoolbox.get_cache_contents()) # doctest: +SKIP
    """
    path = get_save_location()

    if os.path.exists(path):
        return os.listdir(path)

    print("Cache directory does not exist.")
    return None


def read_from_cache(filename: str) -> GridObject:
    """Read a GeoTIFF file from the cache directory and return it as a
    GridObject. The filename should be the name of the GeoTIFF file. Find
    available files by using 'get_cache_contents()'.

    Parameters
    ----------
    filename : str
        Name of the file to be read from the cache directory. Requires the
        whole filename including the extension, like "dem.tif".

    Returns
    -------
    GridObject
        The GridObject generated from the cached GeoTIFF file.

    Example
    -------
    .. plot::

       >>> import topotoolbox
       >>> import matplotlib.pyplot as plt
       >>> dem = topotoolbox.load_dem('bigtujunga')
       >>> dem = topotoolbox.read_from_cache('bigtujunga.tif')
       >>> _= dem.plot()
       >>> plt.show()
    """
    cache_path = os.path.join(get_save_location(), f"{filename}")
    grid_object = read_tif(cache_path)
    return grid_object


def load_opentopography(south: float, north: float, west: float, east: float,
                        api_key: str | None = None, dem_type: str = "SRTMGL3",
                        api_path: str | None = None, overwrite: bool = False,
                        save_path: str | None = None
                        ) -> GridObject:
    """Download a DEM from Open Topography. The DEM is downloaded as a
    GeoTIFF file and saved in the cache directory. The DEM is then
    read into a GridObject. The DEMs come in geographic coordinates (WGS84).
    The API key is required for accessing the OpenTopography API
    (opentopography.org). To avoid having to pass the API key as an argument,
    save it in .opentopography.txt in the working or home directory. To
    overwrite an existing downloaded DEM, use the overwrite parameter. To save
    the DEM to a different location, use the save_path parameter.

    Parameters
    ----------
    south : float
        WGS 84 bounding box south coordinates
    north : float
        WGS 84 bounding box north coordinates
    west : float
        WGS 84 bounding box west coordinates
    east : float
        WGS 84 bounding box east coordinates
    api_key : str | None
        The API key is needed to access the Open Topography API. To get an API
        key, visit opentopography.org, log in, navigate to MyOpenTopoDashboard
        and request a new API key. This argument is used to pass the
        API key directly as a string. By default None.
    api_path : str | None
        The path to a file that contains the API key. If provided, the API key
        will be read from the file. The file should contain only the API key
        without any additional text or formatting. By default None.
    dem_type : str, optional
        Choose one of the available global raster types, by default "SRTMGL3"

        - SRTMGL3 (SRTM GL3 90m)
        - SRTMGL1 (SRTM GL1 30m)
        - SRTMGL1_E (SRTM GL1 Ellipsoidal 30m)
        - AW3D30 (ALOS World 3D 30m)
        - AW3D30_E (ALOS World 3D Ellipsoidal, 30m)
        - SRTM15Plus (Global Bathymetry SRTM15+ V2.1 500m)
        - NASADEM (NASADEM Global DEM)
        - COP30 (Copernicus Global DSM 30m)
        - COP90 (Copernicus Global DSM 90m)
        - EU_DTM (DTM 30m)
        - GEDI_L3 (DTM 1000m)
        - GEBCOIceTopo (Global Bathymetry 500m)
        - GEBCOSubIceTopo (Global Bathymetry 500m)
        - CA_MRDEM_DSM (DSM 30m)
        - CA_MRDEM_DTM (DTM 30m)

    overwrite : bool, optional
        If True cached DEM will be overwritten if it has the same bounds
        and dem_type, by default False
    save_path : str | None, optional
        If provided, the downloaded GeoTIFF will be saved to this path. Like
        this `"path/to/file.tif"` for example. If it is None, files will
        be saved in cache. By default None

    Returns
    -------
    GridObject
        A GridObject generated from the downloaded DEM.

    Raises
    ------
    ValueError
        If the provided DEM type is not valid.
    ConnectionError
        If the API request fails or returns an error.

    Example
    -------
    .. plot::

       >>> import topotoolbox
       >>> import matplotlib.pyplot as plt
       >>> from rasterio import CRS
       >>> dem = topotoolbox.load_opentopography(south=50, north=50.1, west=14.35,
       ...     east=14.6, dem_type="SRTMGL3")
       >>> dem = dem.reproject(CRS.from_epsg(32633), resolution=90)
       >>> _= dem.plot(cmap="terrain")
       >>> plt.show()
    """

    # Check if an API key is provided
    if isinstance(api_key, str):
        api = api_key
    elif isinstance(api_path, str):
        try:
            # since encoding is not predefined, use unspecified encoding
            # pylint: disable=unspecified-encoding
            with open(str(api_path), 'r') as file:
                api = file.read().strip()
        except FileNotFoundError:
            raise FileNotFoundError(
                f"API key file not found: {api_path}") from None
        except PermissionError:
            raise PermissionError(
                f"Cannot read API key file: {api_path}") from None
    else:
        # Since no API key was provided, try defaults:
        if os.path.exists(os.path.abspath('.opentopography.txt')):
            try:
                # pylint: disable=unspecified-encoding
                with open(os.path.abspath(".opentopography.txt"), 'r') as file:
                    api = file.read().strip()
            except (FileNotFoundError, PermissionError):
                pass
        elif os.path.exists(os.path.expanduser('~/.opentopography.txt')):
            try:
                # pylint: disable=unspecified-encoding
                with open(os.path.expanduser("~/.opentopography.txt"), 'r') as file:
                    api = file.read().strip()
            except (FileNotFoundError, PermissionError):
                pass
        else:
            # No API key has been passed or found
            err = ("Neither api_key, api_env or api_path have been provided. "
                   "Default environment variable 'OPENTOPOGRAPHY_API_KEY' or "
                   "file '~/.opentopography.txt' are not set. Use"
                   " api_key='demoapikeyot2022' as the demo key.")
            raise ValueError(err) from None

    if dem_type not in OPEN_TOPO_DATASETS:
        raise ValueError(
            f"Invalid DEM type. Available types are: {OPEN_TOPO_DATASETS}")

    # Assemble the cache path. Create unique/identifiable name for the GeoTIFF
    dem = f"OpenTopo_{south}_{north}_{west}_{east}_{dem_type}"

    # if no save_path is provided, use the cache path
    if save_path is None:
        save_path = os.path.join(get_save_location(), f"{dem}.tif")
    elif os.path.exists(save_path) and not overwrite:
        err = (f"File {save_path} already exists. Use overwrite=True to "
               "overwrite and the file.")
        raise FileExistsError(err)

    if not os.path.exists(save_path) or overwrite:
        url = (f"{OPEN_TOPO_SOURCE}"
               f"?demtype={dem_type}"
               f"&south={south}"
               f"&north={north}"
               f"&west={west}"
               f"&east={east}"
               f"&outputFormat=GTiff"
               f"&API_Key={api}")

        try:
            urlretrieve(url, save_path)
        except urllib.error.HTTPError as e:
            if e.code == 401:
                err = ("Error: 401 - Unauthorized. Check your API key. "
                       "If you don't have one, use this guide to generate one:"
                       " https://opentopography.org/blog/introducing-api-keys"
                       "-access-opentopography-global-datasets")
                raise ConnectionError(err) from None

            error_dict = {
                204: "Bad Data",
                400: "Bad Request",
                500: "Internal Server Error"
            }
            code = e.code
            err = f"Error: {code} - {error_dict.get(code, 'Unknown Error')}"
            raise ConnectionError(err) from None

    grid_obj = read_tif(save_path)
    return grid_obj
